"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateProductionDeviceFunc = void 0;
const alwayscli_1 = require("@alwaysai/alwayscli");
const path_1 = require("path");
const constants_1 = require("../../constants");
const user_1 = require("../user");
const util_1 = require("../../util");
const project_1 = require("../../core/project");
const infrastructure_1 = require("../../infrastructure");
const write_or_validate_device_cfg_file_1 = require("../../core/device/write-or-validate-device-cfg-file");
const device_1 = require("../../core/device");
async function generateProductionDeviceFunc(props) {
    const { name, description } = props;
    await (0, user_1.checkUserIsLoggedInComponent)({ yes: false });
    await (0, project_1.requirePaidPlan)();
    const { username } = await (0, infrastructure_1.CliAuthenticationClient)().getInfo();
    const deviceToAdd = {
        owner: username,
        friendly_name: name,
        host_name: '',
        device_user_name: '',
        description: description || '',
        hardware_ids: '',
        device_hash: '',
        iotKeys: '',
    };
    try {
        const response = await generateDevice(deviceToAdd);
        const outDir = (0, path_1.join)(process.cwd(), response.deviceUUID);
        const cfgDir = (0, path_1.join)(outDir, '.config', 'alwaysai');
        await (0, util_1.runWithSpinner)(async () => {
            const tokenSpawner = (0, util_1.JsSpawner)({ path: cfgDir });
            await (0, device_1.writeTokens)({
                spawner: tokenSpawner,
                tokens: {
                    deviceId: response.deviceUUID,
                    accessToken: response.accessToken,
                    refreshToken: response.refreshToken,
                    idToken: response.idToken,
                },
            });
            await (0, write_or_validate_device_cfg_file_1.writeOrValidateDeviceCfgFile)({
                spawner: tokenSpawner,
                deviceUuid: response.deviceUUID,
            });
            const certSpawner = (0, util_1.JsSpawner)({ path: (0, path_1.join)(cfgDir, 'certificates') });
            await (0, device_1.writeCertificate)({
                spawner: certSpawner,
                certificate: response.iotKeys.certificatePem,
                privateKey: response.iotKeys.keyPair.PrivateKey,
                rootCa: response.iotKeys.rootCA,
            });
        }, [], `Write device credentials to ${outDir}`);
        (0, util_1.echo)(`Created ${name}: device ID=${response.deviceUUID}`);
        return response.deviceUUID;
    }
    catch (error) {
        util_1.logger.error(error);
        throw new alwayscli_1.CliTerseError(`There was an error with creating the device. ${constants_1.PLEASE_REPORT_THIS_ERROR_MESSAGE}`);
    }
}
exports.generateProductionDeviceFunc = generateProductionDeviceFunc;
async function generateDevice(deviceToAdd) {
    let response;
    await (0, util_1.runWithSpinner)(async () => {
        response = await (0, infrastructure_1.addDevice)(deviceToAdd, 'production');
    }, [], 'Create new production device');
    return response;
}
//# sourceMappingURL=generate-production-device.js.map