"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.projectSelectComponent = void 0;
const download = require("download");
const tmp = require("tmp");
const logSymbols = require("log-symbols");
const chalk = require("chalk");
const path_1 = require("path");
const alwayscli_1 = require("@alwaysai/alwayscli");
const fs_extra_1 = require("fs-extra");
const constants_1 = require("../../constants");
const util_1 = require("../../util");
const general_1 = require("../general");
const project_1 = require("../../core/project");
const infrastructure_1 = require("../../infrastructure");
const STARTER_APPS = [
    {
        webName: 'realtime_object_detector',
        purpose: 'Object Detection',
        downloadUrl: 'https://github.com/alwaysai/object-detector/archive/main.zip',
    },
    {
        webName: 'detector_tracker',
        purpose: 'Object Detection with Tracking',
        downloadUrl: 'https://github.com/alwaysai/detector-tracker/archive/main.zip',
    },
    {
        webName: 'image_classifier',
        purpose: 'Image Classification',
        downloadUrl: 'https://github.com/alwaysai/image-classifier/archive/main.zip',
    },
    {
        webName: 'semantic_segmentation_voc',
        purpose: 'Semantic Segmentation',
        downloadUrl: 'https://github.com/alwaysai/semantic-segmentation-voc/archive/main.zip',
    },
    {
        webName: 'realtime_pose_estimator',
        purpose: 'Human Pose Estimation',
        downloadUrl: 'https://github.com/alwaysai/pose-estimator/archive/main.zip',
    },
    {
        webName: 'hello_world_face_detector',
        purpose: 'Face Detector',
        downloadUrl: 'https://github.com/alwaysai/face-detector/archive/main.zip',
    },
    {
        webName: 'instance_segmentation',
        purpose: 'Instance Segmentation',
        downloadUrl: 'https://github.com/alwaysai/instance-segmentation/archive/main.zip',
    },
];
async function projectSelectComponent(props) {
    const { yes, projectUuid } = props;
    if (yes && !projectUuid) {
        throw new alwayscli_1.CliUsageError((0, util_1.RequiredWithYesMessage)('project', undefined));
    }
    const project = projectUuid
        ? await (0, project_1.getProjectByUUID)(projectUuid)
        : await selectFromListOfProjects();
    if (project.starter_app) {
        let appName = project.starter_app;
        // FIXME: Replace legacy "hello_world" starter app with face detector on outdated projects
        if (appName === 'hello_world') {
            appName = 'hello_world_face_detector';
        }
        await downloadStarterApp({ yes, appName });
    }
    return project;
}
exports.projectSelectComponent = projectSelectComponent;
async function selectFromListOfProjects() {
    const authInfo = await (0, infrastructure_1.CliAuthenticationClient)().getInfo();
    const projects = await (0, infrastructure_1.CliRpcClient)().listProjectsUserIsCollaborator({
        user_name: authInfo.username,
    });
    const choices = [
        { title: chalk.green.bold('Create new project'), value: 'new' },
    ];
    projects.forEach((project, index) => {
        choices.push({
            title: project.name,
            value: `${index}`,
        });
    });
    const choice = await (0, util_1.promptForInput)({
        purpose: 'to select a project',
        questions: [
            {
                type: 'select',
                name: 'project',
                message: 'Select a project or create a new project',
                initial: 0,
                choices,
            },
        ],
    });
    if (choice.project === 'new') {
        return await createProject();
    }
    return projects[choice.project];
}
async function createProject() {
    const { username } = await (0, infrastructure_1.CliAuthenticationClient)().getInfo();
    const team = await (0, infrastructure_1.CliRpcClient)().listTeamsUserIsOwner({ user_name: username });
    // TODO: Need to limit input characters to match dashboard implementation
    const input = await (0, util_1.promptForInput)({
        purpose: 'to choose a project name',
        questions: [
            {
                type: 'text',
                name: 'name',
                message: 'Enter a project name:',
            },
        ],
    });
    const name = input.name;
    const choice = await (0, util_1.promptForInput)({
        purpose: 'to select a project type',
        questions: [
            {
                type: 'select',
                name: 'type',
                message: 'How would you like to initialize your project?',
                initial: 0,
                choices: [
                    { title: 'From a Starter App', value: 'starterApp' },
                    { title: 'As an empty app', value: 'empty' },
                ],
            },
        ],
    });
    let starterApp = '';
    switch (choice.type) {
        case 'starterApp': {
            const choices = [];
            STARTER_APPS.forEach((app) => {
                choices.push({
                    title: app.purpose,
                    value: app.webName,
                });
            });
            const choice = await (0, util_1.promptForInput)({
                purpose: 'to select a starter app',
                questions: [
                    {
                        type: 'select',
                        name: 'app',
                        message: 'Select a Starter App',
                        initial: 0,
                        choices,
                    },
                ],
            });
            starterApp = choice.app;
            break;
        }
        case 'empty':
        default:
        // Do nothing
    }
    const project = await (0, infrastructure_1.CliRpcClient)().createProject({
        owner: username,
        team_id: team[0].id,
        name,
        description: '',
        git_url: '',
        starter_app: starterApp,
    });
    return project;
}
async function checkDirIsEmpty(dir) {
    const fileNames = (0, fs_extra_1.readdirSync)(dir);
    return fileNames.length === 0;
}
async function downloadStarterApp(props) {
    const { yes, appName } = props;
    let url = undefined;
    if ((await checkDirIsEmpty(process.cwd())) === false) {
        const proceed = yes ||
            (await (0, general_1.confirmPromptComponent)({
                message: 'Would you like to download the Starter App? Files in this directory may be overwritten!',
            }));
        if (!proceed) {
            (0, util_1.echo)('Skipping download of starter app!');
            return;
        }
    }
    for (const app of STARTER_APPS) {
        if (app.webName === appName) {
            url = app.downloadUrl;
            break;
        }
    }
    if (url === undefined) {
        throw new alwayscli_1.CliTerseError(`Starter App "${appName}" not found!`);
    }
    const tmpDir = tmp.dirSync();
    try {
        await download(url, tmpDir.name, { extract: true });
    }
    catch (err) {
        throw new alwayscli_1.CliTerseError(`Failed to download "${url}", check your internet connection and retry`);
    }
    const fileNames = (0, fs_extra_1.readdirSync)(tmpDir.name);
    if (fileNames.length !== 1 || !fileNames[0]) {
        util_1.logger.error(`Expected application to contain a single directory: ${fileNames}`);
        throw new alwayscli_1.CliTerseError(`Expected application to contain a single directory. ${constants_1.PLEASE_REPORT_THIS_ERROR_MESSAGE}`);
    }
    const destDir = process.cwd();
    const srcDir = (0, path_1.join)(tmpDir.name, fileNames[0]);
    const starterAppFiles = (0, fs_extra_1.readdirSync)(srcDir);
    starterAppFiles.forEach((file) => {
        const src = (0, path_1.join)(srcDir, file);
        const dest = (0, path_1.join)(destDir, file);
        (0, fs_extra_1.copySync)(src, dest);
    });
    (0, fs_extra_1.removeSync)(srcDir);
    tmpDir.removeCallback();
    (0, util_1.echo)(`${logSymbols.success} Download Starter App`);
}
//# sourceMappingURL=project-select-component.js.map