"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const path_1 = require("path");
const tempy = require("tempy");
const yaml_1 = require("yaml");
const paths_1 = require("../../paths");
const update_docker_compose_target_hw_1 = require("./update-docker-compose-target-hw");
const tmpDir = tempy.directory();
const tmpFilePath = (0, path_1.join)(tmpDir, paths_1.DOCKER_COMPOSE_FILE);
describe(update_docker_compose_target_hw_1.updateDockerComposeTargetHw.name, () => {
    let origContents = {};
    beforeEach(() => {
        origContents = {
            services: {
                alwaysai: {
                    build: {
                        context: '.',
                        args: {
                            ALWAYSAI_HW: 'amd64',
                        },
                    },
                    privileged: true,
                },
                server: {
                    build: {
                        context: './server',
                    },
                    network: 'host',
                },
                database: {
                    build: {
                        context: './db',
                        args: {
                            USERNAME: 'db_user',
                        },
                    },
                    network: 'host',
                },
                mq: {
                    image: 'some/image',
                    network: 'host',
                },
            },
        };
    });
    it('skips update due to matching target hardware', async () => {
        const origSerialized = (0, yaml_1.stringify)(origContents);
        (0, fs_1.writeFileSync)(tmpFilePath, origSerialized);
        await (0, update_docker_compose_target_hw_1.updateDockerComposeTargetHw)({ targetHardware: 'amd64', cwd: tmpDir });
        const buffer = (0, fs_1.readFileSync)(tmpFilePath);
        const serialized = buffer.toString();
        const contents = (0, yaml_1.parse)(serialized);
        const expContents = origContents;
        expect(contents).toEqual(expContents);
    });
    it('updates existing target hardware', async () => {
        const origSerialized = (0, yaml_1.stringify)(origContents);
        (0, fs_1.writeFileSync)(tmpFilePath, origSerialized);
        await (0, update_docker_compose_target_hw_1.updateDockerComposeTargetHw)({ targetHardware: 'aarch64', cwd: tmpDir });
        const buffer = (0, fs_1.readFileSync)(tmpFilePath);
        const serialized = buffer.toString();
        const contents = (0, yaml_1.parse)(serialized);
        const expContents = origContents;
        expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'aarch64';
        expect(contents).toEqual(expContents);
    });
    it('Adds additional configurations when switching to jetson', async () => {
        const origSerialized = (0, yaml_1.stringify)(origContents);
        (0, fs_1.writeFileSync)(tmpFilePath, origSerialized);
        await (0, update_docker_compose_target_hw_1.updateDockerComposeTargetHw)({ targetHardware: 'jetson-jp4.6', cwd: tmpDir });
        const buffer = (0, fs_1.readFileSync)(tmpFilePath);
        const serialized = buffer.toString();
        const contents = (0, yaml_1.parse)(serialized);
        const expContents = origContents;
        expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'jetson-jp4.6';
        expContents['services'].alwaysai['runtime'] = 'nvidia';
        expContents['services'].alwaysai['ipc'] = 'host';
        expContents['services'].alwaysai['volumes'] = ['/tmp/argus_socket:/tmp/argus_socket'];
        expect(contents).toEqual(expContents);
    });
    it('Adds additional configurations when switching to jetson with volumes', async () => {
        const origContentsVolumes = origContents;
        origContentsVolumes['services'].alwaysai['volumes'] = ['/home/user/dir:/dir'];
        const origSerialized = (0, yaml_1.stringify)(origContentsVolumes);
        (0, fs_1.writeFileSync)(tmpFilePath, origSerialized);
        await (0, update_docker_compose_target_hw_1.updateDockerComposeTargetHw)({ targetHardware: 'jetson-jp4.6', cwd: tmpDir });
        const buffer = (0, fs_1.readFileSync)(tmpFilePath);
        const serialized = buffer.toString();
        const contents = (0, yaml_1.parse)(serialized);
        const expContents = origContentsVolumes;
        expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'jetson-jp4.6';
        expContents['services'].alwaysai['runtime'] = 'nvidia';
        expContents['services'].alwaysai['ipc'] = 'host';
        expContents['services'].alwaysai['volumes'].push('/tmp/argus_socket:/tmp/argus_socket');
        expect(contents).toEqual(expContents);
    });
    it('Adds additional configurations when switching to jetson with volumes argus in list', async () => {
        const origContentsVolumes = origContents;
        origContentsVolumes['services'].alwaysai['volumes'] = [
            '/home/user/dir:/dir',
            '/tmp/argus_socket:/tmp/argus_socket',
        ];
        const origSerialized = (0, yaml_1.stringify)(origContentsVolumes);
        (0, fs_1.writeFileSync)(tmpFilePath, origSerialized);
        await (0, update_docker_compose_target_hw_1.updateDockerComposeTargetHw)({ targetHardware: 'jetson-jp4.6', cwd: tmpDir });
        const buffer = (0, fs_1.readFileSync)(tmpFilePath);
        const serialized = buffer.toString();
        const contents = (0, yaml_1.parse)(serialized);
        const expContents = origContentsVolumes;
        expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'jetson-jp4.6';
        expContents['services'].alwaysai['runtime'] = 'nvidia';
        expContents['services'].alwaysai['ipc'] = 'host';
        expect(contents).toEqual(expContents);
    });
    it('Adds additional configurations when switching to x86-trt', async () => {
        const origSerialized = (0, yaml_1.stringify)(origContents);
        (0, fs_1.writeFileSync)(tmpFilePath, origSerialized);
        await (0, update_docker_compose_target_hw_1.updateDockerComposeTargetHw)({ targetHardware: 'x86-trt-22.12', cwd: tmpDir });
        const buffer = (0, fs_1.readFileSync)(tmpFilePath);
        const serialized = buffer.toString();
        const contents = (0, yaml_1.parse)(serialized);
        const expContents = origContents;
        expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'x86-trt-22.12';
        expContents['services'].alwaysai['runtime'] = 'nvidia';
        expect(contents).toEqual(expContents);
    });
});
//# sourceMappingURL=update-docker-compose-target-hw.test.js.map