"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const alwayscli_1 = require("@alwaysai/alwayscli");
const run_and_catch_1 = require("@carnesen/run-and-catch");
const add_device_to_project_1 = require("./add-device-to-project");
const constants_1 = require("../../constants");
const infrastructure_1 = require("../../infrastructure");
jest.mock('../../infrastructure');
const mockGetProjectByUUID = jest.fn().mockResolvedValue([]);
const mockGetProjectDevices = jest.fn().mockResolvedValue({});
const mockCheckTeamLimits = jest.fn().mockResolvedValue({});
const mockAddProjectDevice = jest.fn().mockResolvedValue({});
jest.mocked(infrastructure_1.CliRpcClient).mockReturnValue({
    getProjectByUUID: mockGetProjectByUUID,
    getProjectDevices: mockGetProjectDevices,
    checkTeamLimits: mockCheckTeamLimits,
    addProjectDevice: mockAddProjectDevice,
});
function setupMocks(project, projectDevices, teamLimits) {
    mockGetProjectByUUID.mockResolvedValue(project);
    mockGetProjectDevices.mockResolvedValue(projectDevices);
    mockCheckTeamLimits.mockResolvedValue(teamLimits);
}
function checkMocks(project, device, addDevice = true) {
    expect(jest.mocked(infrastructure_1.CliRpcClient)).toBeCalledTimes(2);
    expect(mockGetProjectByUUID).toBeCalledTimes(1);
    expect(mockGetProjectByUUID).toBeCalledWith({ uuid: project.uuid });
    expect(mockGetProjectDevices).toBeCalledTimes(1);
    expect(mockGetProjectDevices).toBeCalledWith({ project_id: project.id });
    expect(mockCheckTeamLimits).toBeCalledTimes(1);
    expect(mockCheckTeamLimits).toBeCalledWith({ team_id: project.team_id });
    if (addDevice) {
        expect(mockAddProjectDevice).toBeCalledTimes(1);
        expect(mockAddProjectDevice).toBeCalledWith({
            device_id: device.id,
            project_id: project.id,
        });
    }
}
describe(add_device_to_project_1.addDeviceToProject.name, () => {
    afterEach(() => {
        jest.clearAllMocks();
    });
    jest.setTimeout(30000);
    const projectId = 'PID17';
    const project = {
        id: 17,
        uuid: projectId,
        team_id: 34,
    };
    const projectDevices = [
        {
            id: 1,
        },
    ];
    it('Device already exists in project', async () => {
        const device = {
            id: 1,
        };
        mockGetProjectByUUID.mockResolvedValue(project);
        mockGetProjectDevices.mockResolvedValue(projectDevices);
        await (0, add_device_to_project_1.addDeviceToProject)({ device, projectId });
        expect(jest.mocked(infrastructure_1.CliRpcClient)).toBeCalledTimes(1);
        expect(mockGetProjectByUUID).toBeCalledTimes(1);
        expect(mockGetProjectByUUID).toBeCalledWith({ uuid: projectId });
        expect(mockGetProjectDevices).toBeCalledTimes(1);
        expect(mockGetProjectDevices).toBeCalledWith({ project_id: project.id });
        expect(mockAddProjectDevice).toBeCalledTimes(0);
    });
    it('New development device with no limit enforcement', async () => {
        const teamLimits = {
            enforce_limits: false,
            dev_devices: {
                limit: 3,
                used: 1,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'development',
        };
        await (0, add_device_to_project_1.addDeviceToProject)({ device, projectId });
        checkMocks(project, device);
    });
    it('New development device without limit', async () => {
        const teamLimits = {
            enforce_limits: true,
            dev_devices: {
                used: 1,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'development',
        };
        await (0, add_device_to_project_1.addDeviceToProject)({ device, projectId });
        checkMocks(project, device);
    });
    it('New development device with limit', async () => {
        const teamLimits = {
            enforce_limits: true,
            dev_devices: {
                limit: 3,
                used: 1,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'development',
        };
        await (0, add_device_to_project_1.addDeviceToProject)({ device, projectId });
        checkMocks(project, device);
    });
    it('New development device at limit', async () => {
        const teamLimits = {
            enforce_limits: true,
            dev_devices: {
                limit: 3,
                used: 3,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'development',
        };
        const exception = await (0, run_and_catch_1.runAndCatch)(add_device_to_project_1.addDeviceToProject, { device, projectId });
        expect(exception.code).toBe(alwayscli_1.CLI_TERSE_ERROR);
        checkMocks(project, device, false);
    });
    it('New production device without limit', async () => {
        const teamLimits = {
            enforce_limits: true,
            prod_devices: {
                used: 8,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'production',
        };
        await (0, add_device_to_project_1.addDeviceToProject)({ device, projectId });
        checkMocks(project, device);
    });
    it('New production device with limit', async () => {
        const teamLimits = {
            enforce_limits: true,
            prod_devices: {
                limit: 10,
                used: 8,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'production',
        };
        await (0, add_device_to_project_1.addDeviceToProject)({ device, projectId });
        checkMocks(project, device);
    });
    it('New production device at limit', async () => {
        const teamLimits = {
            enforce_limits: true,
            prod_devices: {
                limit: 10,
                used: 10,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
            mode: 'production',
        };
        const exception = await (0, run_and_catch_1.runAndCatch)(add_device_to_project_1.addDeviceToProject, { device, projectId });
        expect(exception.code).toBe(alwayscli_1.CLI_TERSE_ERROR);
        checkMocks(project, device, false);
    });
    it('Undefined device mode', async () => {
        const teamLimits = {
            enforce_limits: true,
            prod_devices: {
                limit: 10,
                used: 10,
            },
        };
        setupMocks(project, projectDevices, teamLimits);
        const device = {
            id: 2,
        };
        const exception = await (0, run_and_catch_1.runAndCatch)(add_device_to_project_1.addDeviceToProject, { device, projectId });
        expect(exception.message).toBe(`Invalid device mode (undefined). ${constants_1.PLEASE_REPORT_THIS_ERROR_MESSAGE}`);
        checkMocks(project, device, false);
    });
});
//# sourceMappingURL=add-device-to-project.test.js.map