"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deviceConnect = void 0;
const alwayscli_1 = require("@alwaysai/alwayscli");
const chalk = require("chalk");
const cli_inputs_1 = require("../../cli-inputs");
const user_1 = require("../../components/user");
const util_1 = require("../../util");
const secure_tunnel_1 = require("../../util/secure-tunnel");
const list_1 = require("./list");
// ----------------------------------------------------------------------------
// Local functions
// ----------------------------------------------------------------------------
/**
 * Gets list of connected devices (uuid and device friendly name only)
 * @param {list} deviceList - list connected devices
 * @returns {ConnectedDevice[]} filters out disconnected devices and returns a list of connected devices
 */
function getConnectedDevices(deviceList) {
    const connectedDevices = deviceList.filter((device) => device.status === 'connected');
    return connectedDevices.map((device) => ({
        uuid: device.uuid,
        friendlyName: device.friendly_name,
    }));
}
/**
 * Prompts user to select the device to connect to
 * @param {ConnectedDevice[]} connectedDevices - list of connected devices
 * @returns {string} uuid of the selected device
 */
async function selectDeviceToConnect(connectedDevices) {
    const choices = [{ title: 'None', value: '' }];
    connectedDevices.forEach((device) => {
        choices.push({
            title: `${device.friendlyName}\t(${device.uuid})`,
            value: `${device.uuid}`,
        });
    });
    const choice = await (0, util_1.promptForInput)({
        purpose: 'to select target device to connect',
        questions: [
            {
                type: 'select',
                name: 'targetDevice',
                message: chalk.yellowBright.bold('Select device to ssh to (disconnected devices are not listed)'),
                initial: 0,
                choices,
            },
        ],
    });
    return choice.targetDevice;
}
/**
 * Prompts user to enter user name for the remote host
 * @returns {string} user name for the remote host
 */
async function getRemoteUserLogin() {
    const remoteUserLoginAnswer = await (0, util_1.promptForInput)({
        purpose: 'to set a remote user login',
        questions: [
            {
                type: 'text',
                name: 'login',
                message: chalk.yellowBright.bold('Enter a remote user login'),
            },
        ],
    });
    return remoteUserLoginAnswer.login;
}
// ----------------------------------------------------------------------------
// Main function
// ----------------------------------------------------------------------------
exports.deviceConnect = (0, alwayscli_1.CliLeaf)({
    name: 'connect',
    description: 'Connect to remote devices',
    namedInputs: {
        yes: cli_inputs_1.yesCliInput,
        device: (0, alwayscli_1.CliStringInput)({
            description: 'Device UUID',
        }),
        username: (0, alwayscli_1.CliStringInput)({
            description: 'Device username',
        }),
    },
    hidden: true,
    async action(_, { yes, device, username }) {
        const deletedPorts = (0, secure_tunnel_1.secureTunnelJsonFile)().removeExpiredItems();
        await (0, secure_tunnel_1.removeLocalhostPortsFromKnownHostsFile)(deletedPorts);
        if (yes) {
            if (device === undefined) {
                throw new alwayscli_1.CliUsageError((0, util_1.RequiredWithYesMessage)('device', undefined));
            }
            if (username === undefined) {
                throw new alwayscli_1.CliUsageError((0, util_1.RequiredWithYesMessage)('username', undefined));
            }
        }
        await (0, user_1.checkUserIsLoggedInComponent)({ yes });
        let selectedDeviceUuid = device;
        if (selectedDeviceUuid === undefined) {
            const deviceList = await (0, list_1.getDeviceList)();
            util_1.logger.debug(JSON.stringify(deviceList, null, 4));
            if (deviceList.length === 0) {
                return (0, util_1.echo)(chalk.redBright.bold('No devices found'));
            }
            const connectedDevices = getConnectedDevices(deviceList);
            util_1.logger.debug(`connectedDevices = ${JSON.stringify(connectedDevices, null, 4)}`);
            if (connectedDevices.length === 0) {
                return (0, util_1.echo)(chalk.redBright.bold('No connected devices found'));
            }
            selectedDeviceUuid = await selectDeviceToConnect(connectedDevices);
        }
        util_1.logger.debug(`selectedDevice = ${selectedDeviceUuid}`);
        if (selectedDeviceUuid === '') {
            return (0, util_1.echo)(chalk.redBright.bold('No device selected'));
        }
        const openTunnelData = await (0, secure_tunnel_1.getSourceAccessToken)(selectedDeviceUuid);
        util_1.logger.debug(`openTunnelData = ${JSON.stringify(openTunnelData, null, 4)}`);
        const secureTunnelInfoFromAWS = (0, secure_tunnel_1.processOpenTunnelData)(openTunnelData);
        if (secureTunnelInfoFromAWS.sourceAccessToken.length > 0) {
            const remoteUserLogin = username ? username : await getRemoteUserLogin();
            util_1.logger.debug(`remoteUserLogin = ${remoteUserLogin}`);
            const secureTunnelInfoFromFile = (0, secure_tunnel_1.secureTunnelJsonFile)().getItem(selectedDeviceUuid);
            secureTunnelInfoFromAWS.sshPort =
                Object.keys(secureTunnelInfoFromFile).length > 0
                    ? secureTunnelInfoFromFile.sshPort
                    : (0, secure_tunnel_1.secureTunnelJsonFile)().getNextAvailablePort();
            (0, secure_tunnel_1.secureTunnelJsonFile)().setItem(selectedDeviceUuid, secureTunnelInfoFromAWS);
            const startParams = {
                deviceUuid: selectedDeviceUuid,
                imageName: (0, secure_tunnel_1.getDockerImageName)(),
                sshPort: secureTunnelInfoFromAWS.sshPort,
                sourceAccessToken: secureTunnelInfoFromAWS.sourceAccessToken,
            };
            await (0, secure_tunnel_1.startDockerLocalProxyContainer)(startParams);
            await (0, secure_tunnel_1.startSecureTunnelSshSession)(remoteUserLogin, secureTunnelInfoFromAWS.sshPort);
            await (0, secure_tunnel_1.removeLocalhostPortsFromKnownHostsFile)([secureTunnelInfoFromAWS.sshPort]);
            await (0, secure_tunnel_1.stopDockerLocalProxyContainer)(selectedDeviceUuid);
        }
    },
});
//# sourceMappingURL=connect.js.map