"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const chalk = require("chalk");
const user_1 = require("../../components/user");
const util_1 = require("../../util");
const connect_1 = require("./connect");
const run_and_catch_1 = require("@carnesen/run-and-catch");
const secure_tunnel_1 = require("../../util/secure-tunnel");
const list_1 = require("./list");
const uuid_1 = require("uuid");
// ----------------------------------------------------------------------------
// Global mocks
// ----------------------------------------------------------------------------
jest.mock('../../util/echo');
jest.mock('../../util/prompt-for-input');
jest.mock('../../util/secure-tunnel');
jest.mock('../../components/user');
jest.mock('./list');
// ----------------------------------------------------------------------------
// Test functions
// ----------------------------------------------------------------------------
describe('deviceConnect', () => {
    const userName = 'Camilo';
    const dockerImageName = 'UbuntuDockerImageName';
    const testNextAvailablePort = 8989;
    const devicesListTemplate = [
        {
            id: 561,
            uuid: 'device_uuid_001',
            friendly_name: 'device_friendly_name_001',
            description: 'device_001',
            host_name: '',
            deleted: false,
            status: 'disconnected',
        },
        {
            id: 562,
            uuid: 'device_uuid_002',
            friendly_name: 'device_friendly_name_002',
            description: 'device_002',
            deleted: false,
            status: 'disconnected',
        },
        {
            id: 563,
            uuid: 'device_uuid_003',
            friendly_name: 'device_friendly_name_003',
            description: 'device_003',
            deleted: false,
            status: 'disconnected',
        },
    ];
    const deviceSelectionTemplate = {
        purpose: 'to select target device to connect',
        questions: [
            {
                type: 'select',
                name: 'targetDevice',
                message: chalk.yellowBright.bold('Select device to ssh to (disconnected devices are not listed)'),
                initial: 0,
                choices: [{ title: 'None', value: '' }],
            },
        ],
    };
    const dateTimeEpochNow = Math.floor(Date.now() / 1000);
    const openTunnelDataFromAWSNewTunnel = {
        expiresAt: dateTimeEpochNow + 12 * 3600,
        newTunnelCreated: true,
        sourceAccessToken: 'sourceAccessToken_FromAws_NewTunnel',
        txId: 'txId_FromAws_NewTunnel',
        tunnelId: 'tunnelId_FromAws_NewTunnel',
    };
    const openTunnelDataFromFile = {
        tunnelId: 'tunnelId_FromFile',
        sourceAccessToken: 'sourceAccessToken_FromFile',
        sshPort: 5042,
        expiresAt: dateTimeEpochNow + 9 * 3600,
    };
    const expectedUserLoginPrompt = {
        purpose: 'to set a remote user login',
        questions: [
            {
                type: 'text',
                name: 'login',
                message: chalk.yellowBright.bold('Enter a remote user login'),
            },
        ],
    };
    let mockRemoveExpiredItems;
    let mockRemoveLocalhostPortsFromKnownHostsFile;
    let mockCheckUserIsLoggedInComponent;
    let mockPromptForInput;
    let mockGetSourceAccessToken;
    let mockProcessOpenTunnelData;
    let mockGetItem;
    let mockSetItem;
    let mockGetDockerImageName;
    let mockStartDockerLocalProxyContainer;
    let mockStartSecureTunnelSshSession;
    let mockStopDockerLocalProxyContainer;
    let mockGetNextAvailablePort;
    const setupMocks = (props) => {
        const { expiredPorts, userDevices, userSelectedDevice, openTunnelData, openTunnelDataFromFile, } = props;
        // mock removeExpiredItems
        // ------------------------------------------------------------------------
        mockRemoveExpiredItems = jest.fn();
        mockGetItem = jest.fn();
        mockSetItem = jest.fn();
        mockGetNextAvailablePort = jest.fn();
        mockRemoveExpiredItems.mockReturnValue(expiredPorts);
        mockGetItem.mockReturnValue(openTunnelDataFromFile ? openTunnelDataFromFile : {});
        mockSetItem.mockReturnValue(undefined);
        mockGetNextAvailablePort.mockReturnValue(testNextAvailablePort);
        secure_tunnel_1.secureTunnelJsonFile.mockReturnValue({
            removeExpiredItems: mockRemoveExpiredItems,
            getItem: mockGetItem,
            setItem: mockSetItem,
            getNextAvailablePort: mockGetNextAvailablePort,
        });
        // mock removeLocalhostPortsFromKnownHostsFile
        // ------------------------------------------------------------------------
        mockRemoveLocalhostPortsFromKnownHostsFile = jest.fn();
        mockRemoveLocalhostPortsFromKnownHostsFile.mockResolvedValue(undefined);
        secure_tunnel_1.removeLocalhostPortsFromKnownHostsFile.mockImplementation(mockRemoveLocalhostPortsFromKnownHostsFile);
        // mock checkUserIsLoggedInComponent
        // ------------------------------------------------------------------------
        mockCheckUserIsLoggedInComponent = jest.fn();
        mockCheckUserIsLoggedInComponent.mockResolvedValue(undefined);
        user_1.checkUserIsLoggedInComponent.mockImplementation(mockCheckUserIsLoggedInComponent);
        // mock getDeviceList
        // ------------------------------------------------------------------------
        if (userDevices !== undefined) {
            list_1.getDeviceList.mockResolvedValue(userDevices);
        }
        // mock promptForInput
        // ------------------------------------------------------------------------
        mockPromptForInput = jest.fn();
        if (userDevices !== undefined) {
            mockPromptForInput.mockResolvedValueOnce({ targetDevice: userSelectedDevice });
        }
        mockPromptForInput.mockResolvedValueOnce({ login: userName });
        util_1.promptForInput.mockImplementation(mockPromptForInput);
        // mock getSourceAccessToken
        // ------------------------------------------------------------------------
        mockGetSourceAccessToken = jest.fn();
        mockGetSourceAccessToken.mockResolvedValue(openTunnelData);
        secure_tunnel_1.getSourceAccessToken.mockImplementation(mockGetSourceAccessToken);
        // mock processOpenTunnelData
        // ------------------------------------------------------------------------
        const processedOpenTunnelData = {
            expiresAt: openTunnelData.expiresAt,
            sshPort: 0,
            sourceAccessToken: openTunnelData.sourceAccessToken,
            tunnelId: openTunnelData.tunnelId,
        };
        mockProcessOpenTunnelData = jest.fn();
        mockProcessOpenTunnelData.mockReturnValue(processedOpenTunnelData);
        secure_tunnel_1.processOpenTunnelData.mockImplementation(mockProcessOpenTunnelData);
        // mock getDockerImageName
        // ------------------------------------------------------------------------
        mockGetDockerImageName = jest.fn();
        mockGetDockerImageName.mockReturnValue(dockerImageName);
        secure_tunnel_1.getDockerImageName.mockImplementation(mockGetDockerImageName);
        // mock startDockerLocalProxyContainer
        // ------------------------------------------------------------------------
        mockStartDockerLocalProxyContainer = jest.fn();
        mockStartDockerLocalProxyContainer.mockResolvedValue(undefined);
        secure_tunnel_1.startDockerLocalProxyContainer.mockImplementation(mockStartDockerLocalProxyContainer);
        // mock startSecureTunnelSshSession
        // ------------------------------------------------------------------------
        mockStartSecureTunnelSshSession = jest.fn();
        mockStartSecureTunnelSshSession.mockResolvedValue(undefined);
        secure_tunnel_1.startSecureTunnelSshSession.mockImplementation(mockStartSecureTunnelSshSession);
        // mock stopDockerLocalProxyContainer
        // ------------------------------------------------------------------------
        mockStopDockerLocalProxyContainer = jest.fn();
        mockStopDockerLocalProxyContainer.mockResolvedValue(undefined);
        secure_tunnel_1.stopDockerLocalProxyContainer.mockImplementation(mockStopDockerLocalProxyContainer);
        openTunnelDataFromAWSNewTunnel.txId = (0, uuid_1.v4)();
        openTunnelDataFromAWSNewTunnel.tunnelId = (0, uuid_1.v4)();
        openTunnelDataFromAWSNewTunnel.sourceAccessToken = `${openTunnelDataFromAWSNewTunnel.sourceAccessToken}_${openTunnelDataFromAWSNewTunnel.tunnelId}`;
    };
    const setupChoices = (devicesToSetupArray) => {
        const choicesToSetup = [{ title: 'None', value: '' }];
        for (const device of devicesToSetupArray) {
            choicesToSetup.push({
                title: `${devicesListTemplate[device].friendly_name}\t(${devicesListTemplate[device].uuid})`,
                value: `${devicesListTemplate[device].uuid}`,
            });
        }
        return choicesToSetup;
    };
    beforeEach(() => {
        jest.clearAllMocks();
        jest.resetModules();
    });
    it('should not remove ports from known_hosts file, given no expired tunnels found / no devices found, given API returns empty device array', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const expectedExpiredPorts = [];
        const expectedUserDevices = [];
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: '',
            openTunnelData: openTunnelDataFromAWSNewTunnel,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledWith(expectedExpiredPorts);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(0);
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(0);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(0);
        expect(mockGetItem).toHaveBeenCalledTimes(0);
        expect(mockSetItem).toHaveBeenCalledTimes(0);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(0);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(0);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
    });
    it('should remove expired secure tunnels / No devices found, given API returns empty device array', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const expectedExpiredPorts = [5002, 5042];
        const expectedUserDevices = [];
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: '',
            openTunnelData: openTunnelDataFromAWSNewTunnel,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledWith(expectedExpiredPorts);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(0);
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(0);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(0);
        expect(mockGetItem).toHaveBeenCalledTimes(0);
        expect(mockSetItem).toHaveBeenCalledTimes(0);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(0);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(0);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
    });
    it('3 devices found, but none are connected', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const expectedExpiredPorts = [];
        const expectedUserDevices = JSON.parse(JSON.stringify(devicesListTemplate));
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: '',
            openTunnelData: openTunnelDataFromAWSNewTunnel,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledWith(expectedExpiredPorts);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(0);
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(0);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(0);
        expect(mockGetItem).toHaveBeenCalledTimes(0);
        expect(mockSetItem).toHaveBeenCalledTimes(0);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(0);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(0);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
    });
    it('3 devices found, device 2 is connected, no device selected', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const expectedExpiredPorts = [];
        const expectedUserDevices = JSON.parse(JSON.stringify(devicesListTemplate));
        expectedUserDevices[1].status = 'connected';
        const expectedDeviceSelection = Object.assign({}, deviceSelectionTemplate);
        const expectedChoices = setupChoices([1]);
        expectedDeviceSelection.questions[0].choices = expectedChoices;
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: '',
            openTunnelData: openTunnelDataFromAWSNewTunnel,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledWith(expectedExpiredPorts);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(1);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedDeviceSelection);
        expect(jest.mocked(util_1.echo)).toHaveBeenCalledWith(chalk.redBright.bold('No device selected'));
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(0);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(0);
        expect(mockGetItem).toHaveBeenCalledTimes(0);
        expect(mockSetItem).toHaveBeenCalledTimes(0);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(0);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(0);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(0);
    });
    it('3 devices found, device 1 and 3 are connected, 3rd device selected, sshPort from file', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const expectedExpiredPorts = [];
        const expectedUserDevices = JSON.parse(JSON.stringify(devicesListTemplate));
        expectedUserDevices[0].status = 'connected';
        expectedUserDevices[2].status = 'connected';
        const expectedDeviceSelection = Object.assign({}, deviceSelectionTemplate);
        const expectedChoices = setupChoices([0, 2]);
        expectedDeviceSelection.questions[0].choices = expectedChoices;
        const expectedTunnelData = Object.assign({}, openTunnelDataFromAWSNewTunnel);
        const expectedSecureTunnelDataToWrite = {
            tunnelId: expectedTunnelData.tunnelId,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
            sshPort: openTunnelDataFromFile.sshPort,
            expiresAt: expectedTunnelData.expiresAt,
        };
        const expectedDockerStartParams = {
            deviceUuid: expectedUserDevices[2].uuid,
            imageName: dockerImageName,
            sshPort: openTunnelDataFromFile.sshPort,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
        };
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: expectedUserDevices[2].uuid,
            openTunnelData: expectedTunnelData,
            openTunnelDataFromFile,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(2);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[0][0]).toEqual(expectedExpiredPorts);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[1][0]).toEqual([
            expectedSecureTunnelDataToWrite.sshPort,
        ]);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(2);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedDeviceSelection);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedUserLoginPrompt);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No device selected'));
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(1);
        expect(mockGetSourceAccessToken).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(1);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledWith(expectedTunnelData);
        expect(mockGetItem).toHaveBeenCalledTimes(1);
        expect(mockGetItem).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
        expect(mockSetItem).toHaveBeenCalledTimes(1);
        expect(mockSetItem).toHaveBeenCalledWith(expectedUserDevices[2].uuid, expectedSecureTunnelDataToWrite);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledWith(expectedDockerStartParams);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(1);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledWith(userName, openTunnelDataFromFile.sshPort);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
    });
    it('3 devices found, device 1 and 3 are connected, 3rd device selected, sshPort calculated', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const expectedExpiredPorts = [];
        const expectedUserDevices = JSON.parse(JSON.stringify(devicesListTemplate));
        expectedUserDevices[0].status = 'connected';
        expectedUserDevices[2].status = 'connected';
        const expectedDeviceSelection = Object.assign({}, deviceSelectionTemplate);
        const expectedChoices = setupChoices([0, 2]);
        expectedDeviceSelection.questions[0].choices = expectedChoices;
        const expectedTunnelData = Object.assign({}, openTunnelDataFromAWSNewTunnel);
        const expectedSecureTunnelDataToWrite = {
            tunnelId: expectedTunnelData.tunnelId,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
            sshPort: testNextAvailablePort,
            expiresAt: expectedTunnelData.expiresAt,
        };
        const expectedDockerStartParams = {
            deviceUuid: expectedUserDevices[2].uuid,
            imageName: dockerImageName,
            sshPort: testNextAvailablePort,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
        };
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: expectedUserDevices[2].uuid,
            openTunnelData: expectedTunnelData,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(2);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[0][0]).toEqual(expectedExpiredPorts);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[1][0]).toEqual([
            testNextAvailablePort,
        ]);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(2);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedDeviceSelection);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedUserLoginPrompt);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No device selected'));
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(1);
        expect(mockGetSourceAccessToken).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(1);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledWith(expectedTunnelData);
        expect(mockGetItem).toHaveBeenCalledTimes(1);
        expect(mockGetItem).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
        expect(mockSetItem).toHaveBeenCalledTimes(1);
        expect(mockSetItem).toHaveBeenCalledWith(expectedUserDevices[2].uuid, expectedSecureTunnelDataToWrite);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledWith(expectedDockerStartParams);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(1);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledWith(userName, testNextAvailablePort);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
    });
    it('validates device UUID passed in', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const deviceUuid = '1234';
        const expectedExpiredPorts = [];
        const expectedTunnelData = Object.assign({}, openTunnelDataFromAWSNewTunnel);
        const expectedSecureTunnelDataToWrite = {
            tunnelId: expectedTunnelData.tunnelId,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
            sshPort: testNextAvailablePort,
            expiresAt: expectedTunnelData.expiresAt,
        };
        const expectedDockerStartParams = {
            deviceUuid,
            imageName: dockerImageName,
            sshPort: testNextAvailablePort,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
        };
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userSelectedDevice: deviceUuid,
            openTunnelData: expectedTunnelData,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: deviceUuid, username: undefined }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(2);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[0][0]).toEqual(expectedExpiredPorts);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[1][0]).toEqual([
            testNextAvailablePort,
        ]);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(0);
        expect(mockPromptForInput).toHaveBeenCalledTimes(1);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedUserLoginPrompt);
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(1);
        expect(mockGetSourceAccessToken).toHaveBeenCalledWith(deviceUuid);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(1);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledWith(expectedTunnelData);
        expect(mockGetItem).toHaveBeenCalledTimes(1);
        expect(mockGetItem).toHaveBeenCalledWith(deviceUuid);
        expect(mockSetItem).toHaveBeenCalledTimes(1);
        expect(mockSetItem).toHaveBeenCalledWith(deviceUuid, expectedSecureTunnelDataToWrite);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledWith(expectedDockerStartParams);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(1);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledWith(userName, testNextAvailablePort);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledWith(deviceUuid);
    });
    it('validates device username passed in', async () => {
        // Arrange
        // --------------------------------------------------------------------
        const deviceUsername = 'Yoda';
        const expectedExpiredPorts = [];
        const expectedUserDevices = JSON.parse(JSON.stringify(devicesListTemplate));
        expectedUserDevices[0].status = 'connected';
        expectedUserDevices[2].status = 'connected';
        const expectedDeviceSelection = Object.assign({}, deviceSelectionTemplate);
        const expectedChoices = setupChoices([0, 2]);
        expectedDeviceSelection.questions[0].choices = expectedChoices;
        const expectedTunnelData = Object.assign({}, openTunnelDataFromAWSNewTunnel);
        const expectedSecureTunnelDataToWrite = {
            tunnelId: expectedTunnelData.tunnelId,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
            sshPort: testNextAvailablePort,
            expiresAt: expectedTunnelData.expiresAt,
        };
        const expectedDockerStartParams = {
            deviceUuid: expectedUserDevices[2].uuid,
            imageName: dockerImageName,
            sshPort: testNextAvailablePort,
            sourceAccessToken: expectedTunnelData.sourceAccessToken,
        };
        setupMocks({
            expiredPorts: expectedExpiredPorts,
            userDevices: expectedUserDevices,
            userSelectedDevice: expectedUserDevices[2].uuid,
            openTunnelData: expectedTunnelData,
        });
        // Act
        // --------------------------------------------------------------------
        await connect_1.deviceConnect.action(undefined, { yes: false, device: undefined, username: deviceUsername }, undefined);
        // Assert
        // --------------------------------------------------------------------
        expect(mockRemoveExpiredItems).toHaveBeenCalledTimes(1);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile).toHaveBeenCalledTimes(2);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[0][0]).toEqual(expectedExpiredPorts);
        expect(mockRemoveLocalhostPortsFromKnownHostsFile.mock.calls[1][0]).toEqual([
            expectedSecureTunnelDataToWrite.sshPort,
        ]);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledTimes(1);
        expect(user_1.checkUserIsLoggedInComponent).toHaveBeenCalledWith({ yes: false });
        expect(list_1.getDeviceList).toHaveBeenCalledTimes(1);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No devices found'));
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No connected devices found'));
        expect(mockPromptForInput).toHaveBeenCalledTimes(1);
        expect(mockPromptForInput).toHaveBeenCalledWith(expectedDeviceSelection);
        expect(jest.mocked(util_1.echo)).not.toHaveBeenCalledWith(chalk.redBright.bold('No device selected'));
        expect(mockGetSourceAccessToken).toHaveBeenCalledTimes(1);
        expect(mockGetSourceAccessToken).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledTimes(1);
        expect(mockProcessOpenTunnelData).toHaveBeenCalledWith(expectedTunnelData);
        expect(mockGetItem).toHaveBeenCalledTimes(1);
        expect(mockGetItem).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
        expect(mockSetItem).toHaveBeenCalledTimes(1);
        expect(mockSetItem).toHaveBeenCalledWith(expectedUserDevices[2].uuid, expectedSecureTunnelDataToWrite);
        expect(mockGetDockerImageName).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStartDockerLocalProxyContainer).toHaveBeenCalledWith(expectedDockerStartParams);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledTimes(1);
        expect(mockStartSecureTunnelSshSession).toHaveBeenCalledWith(deviceUsername, testNextAvailablePort);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledTimes(1);
        expect(mockStopDockerLocalProxyContainer).toHaveBeenCalledWith(expectedUserDevices[2].uuid);
    });
    it('requires flags for --yes', async () => {
        await (0, run_and_catch_1.runAndCatch)(connect_1.deviceConnect.action, undefined, { yes: true, device: undefined, username: undefined }, undefined);
        await (0, run_and_catch_1.runAndCatch)(connect_1.deviceConnect.action, undefined, { yes: true, device: '1234', username: undefined }, undefined);
        await (0, run_and_catch_1.runAndCatch)(connect_1.deviceConnect.action, undefined, { yes: true, device: undefined, username: 'abcd' }, undefined);
    });
});
//# sourceMappingURL=connect.test.js.map