import * as logSymbols from 'log-symbols';
import { CliTerseError } from '@alwaysai/alwayscli';

import {
  DockerTargetConfig,
  SshDockerTargetConfig,
  TargetJsonFile,
  TargetJsonFileReturnType,
} from '../../core/app/target-json-file';
import { appCheckComponent } from './app-check-component';
import { DOCKERFILE_STANDALONE, DOCKER_COMPOSE_FILE } from '../../paths';
import {
  checkForDockerComposeComponent,
  writeStandaloneDockerfileComponent,
  findOrWriteDockerComposeFileComponent,
} from '../docker';
import { echo, JsSpawner, copyFile } from '../../util';
import { requirePaidPlan } from '../../core/project';
import {
  dockerComposeLogs,
  dockerComposeDown,
  dockerComposeBuild,
  dockerComposeUp,
} from '../../util/docker/docker-compose-cmd';
import { getDockerComposeCmdForApp, updateDockerComposeTargetHw } from '../../core/app';

export async function appDeployLinuxAndRemoteDevice(props: {
  yes: boolean;
  generateDockerCompose: boolean;
  logs: boolean;
  stop: boolean;
  targetJson: TargetJsonFileReturnType;
  targetConfig: SshDockerTargetConfig | DockerTargetConfig;
}) {
  const { yes, generateDockerCompose, logs, stop, targetJson, targetConfig } = props;
  const targetHostSpawner = targetJson.readHostSpawner();
  const { targetHardware } = targetConfig;
  if (logs) {
    await dockerComposeLogs(targetHostSpawner);
    return;
  }
  if (stop) {
    await dockerComposeDown(targetHostSpawner);
    echo(`${logSymbols.success} App stopped.`);
    return;
  }

  await checkForDockerComposeComponent(targetHostSpawner);

  await writeStandaloneDockerfileComponent();

  if (targetConfig.targetProtocol === 'ssh+docker:') {
    await copyFile(JsSpawner(), targetHostSpawner, DOCKERFILE_STANDALONE);
  }

  const found = await findOrWriteDockerComposeFileComponent({
    yes,
    cmd: await getDockerComposeCmdForApp({ targetHardware }),
  });
  if (found) {
    await updateDockerComposeTargetHw({ targetHardware });
  }

  if (generateDockerCompose) {
    return;
  }

  if (targetConfig.targetProtocol === 'ssh+docker:') {
    await copyFile(JsSpawner(), targetHostSpawner, DOCKER_COMPOSE_FILE);
  }

  await dockerComposeBuild(targetHostSpawner);
  await dockerComposeUp(targetHostSpawner);

  echo(`${logSymbols.success} App deployed.`);
}

export async function appDeployComponent(props: {
  yes: boolean;
  generateDockerCompose: boolean;
  logs: boolean;
  stop: boolean;
}) {
  const { yes, generateDockerCompose, logs, stop } = props;
  await appCheckComponent();
  await requirePaidPlan();
  const targetJson = TargetJsonFile();
  const targetConfig = targetJson.read();

  switch (targetConfig.targetProtocol) {
    case 'docker:':
    case 'ssh+docker:': {
      await appDeployLinuxAndRemoteDevice({
        yes,
        generateDockerCompose,
        logs,
        stop,
        targetJson,
        targetConfig,
      });
      break;
    }

    default: {
      throw new CliTerseError(`Deploy is not supported for native installations.`);
    }
  }
}
