import { CliTerseError, CliUsageError } from '@alwaysai/alwayscli';
import { checkProcessorBrand } from './destination-prompt-component';
import { targetPathPromptComponent } from './target-path-prompt-component';
import { targetProtocolPromptComponent } from './target-protocol-prompt-component';
import { writeTargetJsonFileComponent } from './write-target-json-file-component';
import {
  DOCKER_IMAGE_ID_INITIAL_VALUE,
  ALWAYSAI_CLI_EXECUTABLE_NAME,
} from '../../../constants';
import {
  deviceCheckAndUpdateComponent,
  deviceSelectComponent,
  getTargetHostnameComponent,
} from '../../device';
import { checkForDockerComponent } from '../../docker';
import { RequiredWithYesMessage } from '../../../util';
import {
  getTargetHardwareType,
  TargetHardware,
  TargetJsonFile,
  TargetPathDefaultValue,
  TargetProtocol,
} from '../../../core/app';
import { addDeviceToProject, ProjectJsonFile } from '../../../core/project';
import { findOrWritePrivateKeyFileComponent } from '../../general';

export async function targetJsonComponent(props: {
  yes: boolean;
  targetProtocol?: TargetProtocol;
  targetHardware?: TargetHardware;
  targetHostname?: string;
  targetPath?: string;
  deviceId?: string;
}) {
  const { yes } = props;
  const currentTargetJson = TargetJsonFile();
  const currentTargetProtocol = currentTargetJson.readTargetProtocolSafe();
  let targetProtocol;
  if (yes) {
    targetProtocol = props.targetProtocol || currentTargetProtocol;
    if (!targetProtocol) {
      throw new CliUsageError(RequiredWithYesMessage('protocol', undefined));
    }
    if (targetProtocol === 'native:' || targetProtocol === 'docker:') {
      if (!(await checkProcessorBrand())) {
        throw new CliUsageError(
          'edgeIQ is not supported on devices with Atom/Celeron processors.',
        );
      }
    }
  } else if (!props.targetProtocol) {
    targetProtocol = await targetProtocolPromptComponent({
      prevTargetProtocol: currentTargetProtocol,
    });
  } else {
    targetProtocol = props.targetProtocol;
  }

  switch (targetProtocol) {
    case 'native:': {
      writeTargetJsonFileComponent({
        targetProtocol,
      });
      break;
    }

    case 'docker:': {
      await checkForDockerComponent();

      const targetHardware = props.targetHardware || (await getTargetHardwareType({}));

      writeTargetJsonFileComponent({
        targetProtocol,
        dockerImageId: DOCKER_IMAGE_ID_INITIAL_VALUE,
        targetHardware,
      });
      break;
    }

    case 'ssh+docker:': {
      await findOrWritePrivateKeyFileComponent({ yes });
      const projectJsonFile = ProjectJsonFile();
      const projectConfig = projectJsonFile.read().project;
      if (!projectConfig) {
        throw new CliTerseError(
          `App must have a project associated with it. Run '${ALWAYSAI_CLI_EXECUTABLE_NAME} app configure' to select a project.`,
        );
      }
      const projectId = projectConfig.id;
      const device = await deviceSelectComponent({
        yes,
        deviceId: props.deviceId,
      });

      const targetHostname = await getTargetHostnameComponent({
        yes,
        device,
        targetHostname: props.targetHostname,
      });
      if (targetHostname === undefined) {
        throw new CliTerseError(
          'Cannot access selected device! The device hostname is unknown.',
        );
      }

      await deviceCheckAndUpdateComponent({ yes, targetHostname, device });

      await addDeviceToProject({ device, projectId });

      await checkForDockerComponent({ targetHostname });

      let targetPath =
        props.targetPath ||
        currentTargetJson.readFieldSafe({ name: 'targetPath' }) ||
        TargetPathDefaultValue();
      if (!yes) {
        targetPath = await targetPathPromptComponent({
          targetHostname,
          targetPath,
        });
      }

      const targetHardware =
        props.targetHardware || (await getTargetHardwareType({ targetHostname }));

      TargetJsonFile().write({
        targetProtocol,
        targetHostname,
        targetPath,
        dockerImageId: DOCKER_IMAGE_ID_INITIAL_VALUE,
        targetHardware,
        deviceId: device.uuid,
      });
      break;
    }

    default:
      throw new Error('Unsupported protocol');
  }
}
