import { CLI_TERSE_ERROR } from '@alwaysai/alwayscli';
import { runAndCatch } from '@carnesen/run-and-catch';
import { promptForInput, SshSpawner } from '../../util';
import { deviceAddComponent } from './device-add-component';
import { getTargetHardwareUuid } from '../../core/app';
import { getUpgradeUrl } from '../../core/project';
import { CliRpcClient, addDevice, CliAuthenticationClient } from '../../infrastructure';
import { targetHostnameCheckComponent } from './target-hostname-check-component';

const CALLED_OPEN = 'Called open';
jest.mock('open', () => {
  return jest.fn().mockImplementation(() => {
    throw new Error(CALLED_OPEN);
  });
});

jest.mock('@sentry/node');

jest.mock('../../infrastructure');
const mockGetInfo = jest.fn().mockResolvedValue({});
jest
  .mocked(CliAuthenticationClient as jest.Mock)
  .mockReturnValue({ getInfo: mockGetInfo });
const mockListTeamsUserIsMember = jest.fn().mockResolvedValue([]);
const mockCheckTeamLimits = jest.fn().mockResolvedValue({});
jest.mocked(CliRpcClient as jest.Mock).mockReturnValue({
  listTeamsUserIsMember: mockListTeamsUserIsMember,
  checkTeamLimits: mockCheckTeamLimits,
});

jest.mock('./target-hostname-check-component');
jest.mock('../../core/app');
jest.mock('../../core/project');

jest.mock('../../util/spawner');
const mockRun = jest.fn();
jest.mocked(SshSpawner as jest.Mock).mockReturnValue({ run: mockRun });

jest.mock('../../util/prompt-for-input');

function setupMocks({
  teamId,
  teamLimits,
  friendlyName,
  url,
  hostname,
  hardwareId,
  username,
  deviceUsername,
}) {
  mockGetInfo.mockResolvedValue({ username });
  mockListTeamsUserIsMember.mockResolvedValue([{ id: teamId }]);
  mockCheckTeamLimits.mockResolvedValue(teamLimits);
  jest.mocked(promptForInput).mockResolvedValueOnce({ name: friendlyName });
  jest.mocked(getUpgradeUrl).mockResolvedValue(url);
  jest.mocked(targetHostnameCheckComponent).mockResolvedValue(hostname);
  jest.mocked(getTargetHardwareUuid).mockResolvedValue(hardwareId);
  mockRun.mockResolvedValue(deviceUsername);
  jest
    .mocked(addDevice as jest.Mock)
    .mockResolvedValue({ deviceId: 10, deviceUUID: '20' });
}

function checkMocks({
  teamId,
  username,
  friendlyName,
  hostname,
  deviceUsername,
  hardwareId,
  deviceMode,
  hitLimit = false,
}) {
  expect(jest.mocked(CliRpcClient)).toBeCalledTimes(1);
  expect(mockListTeamsUserIsMember).toBeCalledTimes(1);
  expect(mockListTeamsUserIsMember).toBeCalledWith({ user_name: username });
  expect(mockCheckTeamLimits).toBeCalledTimes(1);
  expect(mockCheckTeamLimits).toBeCalledWith({ team_id: teamId });
  expect(mockGetInfo).toBeCalledTimes(1);
  if (hitLimit) {
    expect(jest.mocked(targetHostnameCheckComponent)).toBeCalledTimes(0);
    expect(jest.mocked(getTargetHardwareUuid)).toBeCalledTimes(0);
    expect(jest.mocked(addDevice)).toBeCalledTimes(0);
  } else {
    expect(jest.mocked(promptForInput)).toBeCalledTimes(1);
    expect(jest.mocked(targetHostnameCheckComponent)).toBeCalledTimes(1);
    expect(jest.mocked(getTargetHardwareUuid)).toBeCalledTimes(1);
    expect(jest.mocked(addDevice)).toBeCalledTimes(1);
    expect(jest.mocked(addDevice)).toBeCalledWith(
      {
        owner: username,
        friendly_name: friendlyName,
        host_name: hostname,
        device_user_name: deviceUsername,
        description: '',
        hardware_ids: hardwareId,
        device_hash: '',
        iotKeys: '',
      },
      deviceMode,
    );
  }
}

describe(deviceAddComponent.name, () => {
  afterEach(() => {
    jest.clearAllMocks();
  });
  jest.setTimeout(30000);
  const username = 'Darth Vader';
  const teamId = 7;
  const deviceMode = 'development';
  const url = 'https://test';

  it('--yes not supported', async () => {
    const exception = await runAndCatch(deviceAddComponent, {
      yes: true,
    });
    expect(exception.code).toBe(CLI_TERSE_ERROR);
    expect(exception.message).toMatch(/Adding a new device is not supported with --yes/i);
  });

  it("Team doesn't have device limits", async () => {
    const friendlyName = 'R2D2';
    const hostname = 'r2d2@beepboop';
    const hardwareId = "you're my only hope";
    const deviceUsername = 'r2d2';
    setupMocks({
      teamId,
      teamLimits: {
        enforce_limits: false,
      },
      friendlyName,
      url,
      hostname,
      hardwareId,
      username,
      deviceUsername,
    });

    await deviceAddComponent({ yes: false });

    checkMocks({
      teamId,
      username,
      friendlyName,
      hostname,
      deviceUsername,
      deviceMode,
      hardwareId,
    });
  });

  it('Team is below device limit', async () => {
    const friendlyName = 'R2D2';
    const devices = { limit: 11, used: 3 };
    const hostname = 'r2d2@beepboop';
    const hardwareId = "you're my only hope";
    const deviceUsername = 'r2d2';
    setupMocks({
      teamId,
      teamLimits: {
        enforce_limits: true,
        dev_devices: devices,
        prod_devices: devices,
      },
      friendlyName,
      url,
      hostname,
      hardwareId,
      username,
      deviceUsername,
    });

    await deviceAddComponent({ yes: false });

    checkMocks({
      teamId,
      username,
      friendlyName,
      hostname,
      deviceUsername,
      deviceMode,
      hardwareId,
    });
  });

  it('Team is at device limit', async () => {
    const friendlyName = 'R2D2';
    const devices = { limit: 12, used: 12 };
    const hostname = 'r2d2@beepboop';
    const hardwareId = "you're my only hope";
    const deviceUsername = 'r2d2';
    setupMocks({
      teamId,
      teamLimits: {
        enforce_limits: true,
        dev_devices: devices,
        prod_devices: devices,
      },
      friendlyName,
      url,
      hostname,
      hardwareId,
      username,
      deviceUsername,
    });

    const exception = await runAndCatch(deviceAddComponent, {
      yes: false,
    });
    expect(exception.message).toBe('No available development devices left.');

    checkMocks({
      teamId,
      username,
      friendlyName,
      hostname,
      deviceUsername,
      hardwareId,
      deviceMode,
      hitLimit: true,
    });
  });
});
