import { CLI_USAGE_ERROR } from '@alwaysai/alwayscli';
import { runAndCatch } from '@carnesen/run-and-catch';
import { deviceAddComponent } from './device-add-component';
import { deviceSelectComponent } from './device-select-component';
import { promptForInput } from '../../util';
import {
  CliRpcClient,
  CliAuthenticationClient,
  AaiDevice,
  getDeviceByUuid,
} from '../../infrastructure';

jest.mock('../../infrastructure');

const mockGetUserDevices = jest.fn().mockResolvedValue({});
const mockCheckTeamLimits = jest.fn().mockResolvedValue({});
const mockCheckTeamsUserIsOwner = jest.fn().mockResolvedValue({});
jest.mocked(CliRpcClient as jest.Mock).mockReturnValue({
  getUserDevices: mockGetUserDevices,
  checkTeamLimits: mockCheckTeamLimits,
  listTeamsUserIsOwner: mockCheckTeamsUserIsOwner,
});

const mockGetInfo = jest.fn().mockResolvedValue({});
jest.mocked(CliAuthenticationClient as jest.Mock).mockReturnValue({
  getInfo: mockGetInfo,
});

jest.mock('../../util/prompt-for-input');
jest.mock('./device-add-component');

function createDevice(props: { friendlyName: string; hwId?: string; mode?: string }) {
  const { friendlyName, hwId, mode = 'production' } = props;
  const id = Math.random();
  return {
    owner: '',
    friendly_name: friendlyName,
    host_name: `${friendlyName}.local`,
    device_user_name: '',
    description: '',
    hardware_ids: hwId || '1234',
    device_hash: '',
    id,
    uuid: `${id}`,
    mode,
    cognito_device_key: '',
    deleted: false,
    created_at: '',
    updated_at: '',
  } as AaiDevice;
}

function setupMocksForDeviceSelectComponentPrompt(props: {
  username;
  userDevices: AaiDevice[];
  selectedDevice: AaiDevice;
  targetDevice: string | number;
}) {
  const { username, userDevices, selectedDevice, targetDevice = 'new' } = props;
  mockGetInfo.mockResolvedValue({ username });
  mockGetUserDevices.mockResolvedValue(userDevices);
  jest.mocked(promptForInput).mockResolvedValue({
    targetDevice,
  });
  if (targetDevice === 'new') {
    jest.mocked(deviceAddComponent as jest.Mock).mockResolvedValue(selectedDevice);
  }
}

function checkMocksForDeviceSelectComponentPrompt({ isNew }: { isNew: boolean }) {
  expect(mockGetInfo).toBeCalledTimes(1);
  expect(mockGetUserDevices).toBeCalledTimes(1);
  expect(jest.mocked(promptForInput)).toBeCalledTimes(1);
  if (isNew) {
    expect(jest.mocked(deviceAddComponent)).toBeCalledTimes(1);
  } else {
    expect(jest.mocked(deviceAddComponent)).toBeCalledTimes(0);
  }
}

describe(deviceSelectComponent.name, () => {
  afterEach(() => {
    jest.clearAllMocks();
  });
  jest.setTimeout(30000);

  describe('--yes', () => {
    test('without deviceId arg', async () => {
      const exception = await runAndCatch(deviceSelectComponent, {
        yes: true,
      });

      expect(exception.code).toBe(CLI_USAGE_ERROR);
    });

    it('with deviceId arg', async () => {
      const device = createDevice({ friendlyName: 'mainsail' });
      jest.mocked(getDeviceByUuid).mockResolvedValue(device);

      const out = await deviceSelectComponent({
        yes: true,
        deviceId: device.uuid,
      });

      expect(jest.mocked(getDeviceByUuid)).toBeCalledTimes(1);
      expect(jest.mocked(getDeviceByUuid).mock.calls[0][0]).toEqual({
        uuid: device.uuid,
      });
      expect(out).toBe(device);
    });
  });

  describe('prompt for device', () => {
    test('select existing device successfully', async () => {
      const username = 'Chewbacca';
      const hwId = '5678';
      const device0 = createDevice({ friendlyName: 'stormjib', hwId: '1234' });
      const device1 = createDevice({ friendlyName: 'windseeker', hwId });
      const devices = [device0, device1];
      setupMocksForDeviceSelectComponentPrompt({
        username,
        userDevices: devices,
        selectedDevice: device1,
        targetDevice: 1,
      });

      const out = await deviceSelectComponent({
        yes: false,
      });

      checkMocksForDeviceSelectComponentPrompt({
        isNew: false,
      });
      expect(out).toBe(device1);
    });
  });
});
