import { createWriteStream } from 'fs';
import { basename, join } from 'path';
import nodeFetch from 'node-fetch';
import { CliTerseError } from '@alwaysai/alwayscli';
import { checkUserIsLoggedInComponent } from '../user';
import { getProjectReleases } from '../../core/release/get-release-versions';
import { logger, runWithSpinner } from '../../util';
import { fetchAppReleaseHistory, getReleaseURL } from '../../infrastructure';
import { validateIsUserProject } from '../../core/project';

export async function fetchAppReleaseHistoryComponent(props: {
  yes: boolean;
  project: string;
}) {
  const { yes, project } = props;

  await checkUserIsLoggedInComponent({ yes });

  await validateIsUserProject({ project });
  return await fetchAppReleaseHistory(project);
}

export async function appReleasePullComponent(props: {
  yes: boolean;
  project: string;
  releaseHash?: string;
}) {
  const { yes, project, releaseHash } = props;

  await checkUserIsLoggedInComponent({ yes });
  await validateIsUserProject({ project });

  // verify the project has been published (also verifies project ownership)
  const releaseHistory = await fetchAppReleaseHistory(project);
  if (releaseHistory.length === 0) {
    throw new CliTerseError('This project has not been published!');
  }

  // verify valid release hash, if provided
  if (releaseHash !== undefined) {
    const releases = await getProjectReleases({ releaseHistory });
    if (!releases.includes(releaseHash)) {
      throw new CliTerseError(
        `The release hash you specified ${releaseHash} is invalid for this project!`,
      );
    }
  }

  await runWithSpinner(
    async () => {
      let key;
      let presignedAppUrl;
      try {
        const results = await getReleaseURL(project, releaseHash);
        key = results.key;
        presignedAppUrl = results.presignedAppUrl;
      } catch (error) {
        throw new CliTerseError('An error occurred retrieving project.');
      }
      const response = await nodeFetch(presignedAppUrl);
      if (response.status !== 200) {
        // If the URL is invalid; I think we shouldn't get here with the new changes
        logger.error(
          `Status Code: ${response.status}, ${response.statusText}. Response: ${response.body}`,
        );
        throw new CliTerseError(
          'Please verify you have specified the project and release version properly.',
        );
      }

      const localDest = join(process.cwd(), `${basename(key)}`);
      const stream = response.body.pipe(createWriteStream(localDest));

      await new Promise((resolve, reject) => {
        stream.on('finish', resolve);
        stream.on('error', reject);
      });
    },
    [],
    'Retrieving application package',
  );

  return releaseHash;
}
