import * as t from 'io-ts';
import { CliTerseError } from '@alwaysai/alwayscli';
import { keyMirror, SshSpawner, JsSpawner, getL4TVersion } from '../../util';

export const TargetHardware = keyMirror({
  'jetson-jp4.6': null,
  'jetson-jp4.6.1': null,
  'jetson-jp4.6.2': null,
  'jetson-jp4.6.3': null,
  'jetson-jp4.6.4': null,
  'jetson-jp5.1': null,
  armv7hf: null,
  aarch64: null,
  amd64: null,
  'x86-trt-22.12': null,
  'x86-trt-23.02': null,
  'eyecloud-amd64': null,
  'eyecloud-armv7hf': null,
  'hailo-amd64': null,
});

export type TargetHardware = keyof typeof TargetHardware;
export const targetHardware = t.keyof(TargetHardware);
export const TARGET_HARDWARE_OPTIONS = Object.keys(TargetHardware) as TargetHardware[];

export async function getTargetHardwareType(props: {
  targetHostname?: string;
}): Promise<TargetHardware> {
  const { targetHostname } = props;
  const spawner = targetHostname
    ? SshSpawner({ targetHostname })
    : JsSpawner({ path: '.' });

  let targetHardware: TargetHardware;

  const arch = await spawner.run({
    exe: 'uname',
    args: ['-m'],
  });
  const kernel = await spawner.run({
    exe: 'uname',
    args: ['-r'],
  });

  if (kernel.includes('tegra') || kernel.includes('l4t')) {
    const l4tVersion = await getL4TVersion({ spawner });
    if (l4tVersion === null) {
      throw new CliTerseError('JetPack version not found!');
    }
    switch (l4tVersion) {
      case '32.6.1':
        targetHardware = 'jetson-jp4.6';
        break;
      case '32.7.1':
        targetHardware = 'jetson-jp4.6.1';
        break;
      case '32.7.2':
        targetHardware = 'jetson-jp4.6.2';
        break;
      case '32.7.3':
        targetHardware = 'jetson-jp4.6.3';
        break;
      case '32.7.4':
        targetHardware = 'jetson-jp4.6.4';
        break;
      case '35.2.1':
        targetHardware = 'jetson-jp5.1';
        break;
      default:
        throw new CliTerseError(`L4T version ${l4tVersion} not supported!`);
    }
  } else if (arch.includes('armv7')) {
    targetHardware = 'armv7hf';
  } else if (arch.includes('aarch64') || arch.includes('armv8')) {
    targetHardware = 'aarch64';
  } else if (arch.includes('amd64') || arch.includes('x86_64')) {
    targetHardware = 'amd64';
  } else {
    throw new CliTerseError(`Target hardware (${arch}) not supported!`);
  }

  return targetHardware;
}
