import { readFileSync, writeFileSync } from 'fs';
import { join } from 'path';
import * as tempy from 'tempy';
import { parse, stringify } from 'yaml';
import { DOCKER_COMPOSE_FILE } from '../../paths';

import { updateDockerComposeTargetHw } from './update-docker-compose-target-hw';

const tmpDir = tempy.directory();
const tmpFilePath = join(tmpDir, DOCKER_COMPOSE_FILE);

describe(updateDockerComposeTargetHw.name, () => {
  let origContents = {};
  beforeEach(() => {
    origContents = {
      services: {
        alwaysai: {
          build: {
            context: '.',
            args: {
              ALWAYSAI_HW: 'amd64',
            },
          },
          privileged: true,
        },
        server: {
          build: {
            context: './server',
          },
          network: 'host',
        },
        database: {
          build: {
            context: './db',
            args: {
              USERNAME: 'db_user',
            },
          },
          network: 'host',
        },
        mq: {
          image: 'some/image',
          network: 'host',
        },
      },
    };
  });
  it('skips update due to matching target hardware', async () => {
    const origSerialized = stringify(origContents);
    writeFileSync(tmpFilePath, origSerialized);

    await updateDockerComposeTargetHw({ targetHardware: 'amd64', cwd: tmpDir });

    const buffer = readFileSync(tmpFilePath);
    const serialized = buffer.toString();
    const contents = parse(serialized);
    const expContents = origContents;
    expect(contents).toEqual(expContents);
  });
  it('updates existing target hardware', async () => {
    const origSerialized = stringify(origContents);
    writeFileSync(tmpFilePath, origSerialized);

    await updateDockerComposeTargetHw({ targetHardware: 'aarch64', cwd: tmpDir });

    const buffer = readFileSync(tmpFilePath);
    const serialized = buffer.toString();
    const contents = parse(serialized);
    const expContents = origContents;
    expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'aarch64';
    expect(contents).toEqual(expContents);
  });
  it('Adds additional configurations when switching to jetson', async () => {
    const origSerialized = stringify(origContents);
    writeFileSync(tmpFilePath, origSerialized);

    await updateDockerComposeTargetHw({ targetHardware: 'jetson-jp4.6', cwd: tmpDir });

    const buffer = readFileSync(tmpFilePath);
    const serialized = buffer.toString();
    const contents = parse(serialized);
    const expContents = origContents;
    expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'jetson-jp4.6';
    expContents['services'].alwaysai['runtime'] = 'nvidia';
    expContents['services'].alwaysai['ipc'] = 'host';
    expContents['services'].alwaysai['volumes'] = ['/tmp/argus_socket:/tmp/argus_socket'];
    expect(contents).toEqual(expContents);
  });
  it('Adds additional configurations when switching to jetson with volumes', async () => {
    const origContentsVolumes = origContents;
    origContentsVolumes['services'].alwaysai['volumes'] = ['/home/user/dir:/dir'];
    const origSerialized = stringify(origContentsVolumes);
    writeFileSync(tmpFilePath, origSerialized);

    await updateDockerComposeTargetHw({ targetHardware: 'jetson-jp4.6', cwd: tmpDir });

    const buffer = readFileSync(tmpFilePath);
    const serialized = buffer.toString();
    const contents = parse(serialized);
    const expContents = origContentsVolumes;
    expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'jetson-jp4.6';
    expContents['services'].alwaysai['runtime'] = 'nvidia';
    expContents['services'].alwaysai['ipc'] = 'host';
    expContents['services'].alwaysai['volumes'].push(
      '/tmp/argus_socket:/tmp/argus_socket',
    );
    expect(contents).toEqual(expContents);
  });
  it('Adds additional configurations when switching to jetson with volumes argus in list', async () => {
    const origContentsVolumes = origContents;
    origContentsVolumes['services'].alwaysai['volumes'] = [
      '/home/user/dir:/dir',
      '/tmp/argus_socket:/tmp/argus_socket',
    ];
    const origSerialized = stringify(origContentsVolumes);
    writeFileSync(tmpFilePath, origSerialized);

    await updateDockerComposeTargetHw({ targetHardware: 'jetson-jp4.6', cwd: tmpDir });

    const buffer = readFileSync(tmpFilePath);
    const serialized = buffer.toString();
    const contents = parse(serialized);
    const expContents = origContentsVolumes;
    expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'jetson-jp4.6';
    expContents['services'].alwaysai['runtime'] = 'nvidia';
    expContents['services'].alwaysai['ipc'] = 'host';
    expect(contents).toEqual(expContents);
  });
  it('Adds additional configurations when switching to x86-trt', async () => {
    const origSerialized = stringify(origContents);
    writeFileSync(tmpFilePath, origSerialized);

    await updateDockerComposeTargetHw({ targetHardware: 'x86-trt-22.12', cwd: tmpDir });

    const buffer = readFileSync(tmpFilePath);
    const serialized = buffer.toString();
    const contents = parse(serialized);
    const expContents = origContents;
    expContents['services'].alwaysai.build.args.ALWAYSAI_HW = 'x86-trt-22.12';
    expContents['services'].alwaysai['runtime'] = 'nvidia';
    expect(contents).toEqual(expContents);
  });
});
