import { Spawner } from '../../util';
import { writeOrValidateDeviceCfgFile } from './write-or-validate-device-cfg-file';
import { runAndCatch } from '@carnesen/run-and-catch';

function getSpawner(args: { fileExists?: boolean; readFileOut?: string }) {
  const { fileExists, readFileOut } = args;
  const spawner = {
    exists: jest.fn().mockResolvedValue(fileExists === true ? true : false),
    mkdirp: jest.fn().mockResolvedValue({}),
    writeFile: jest.fn().mockResolvedValue({}),
    readFile: jest.fn().mockResolvedValue(readFileOut),
  } as any as Spawner;
  return spawner;
}

describe('Test writeOrValidateDeviceCfgFile', () => {
  test("file doesn't exist", async () => {
    const spawner = getSpawner({});
    const deviceUuid = '1234';
    await writeOrValidateDeviceCfgFile({ spawner, deviceUuid });

    expect(spawner.exists).toBeCalled();
    expect(spawner.mkdirp).toBeCalled();
    expect(spawner.writeFile).toBeCalled();
  });

  test('File exists but invalid json', async () => {
    const readFileOut = 'not valid json';
    const spawner = getSpawner({ fileExists: true, readFileOut });
    const deviceUuid = '1234';

    await runAndCatch(writeOrValidateDeviceCfgFile, { spawner, deviceUuid });

    expect(spawner.exists).toBeCalled();
    expect(spawner.readFile).toBeCalled();
  });

  test('File exists but invalid contents', async () => {
    const readFileOut = JSON.stringify({
      notAField: 'not a field',
    });
    const spawner = getSpawner({ fileExists: true, readFileOut });
    const deviceUuid = '1234';

    await runAndCatch(writeOrValidateDeviceCfgFile, { spawner, deviceUuid });

    expect(spawner.exists).toBeCalled();
    expect(spawner.readFile).toBeCalled();
  });

  test('File exists with mismatching device UUID', async () => {
    const readFileOut = JSON.stringify({
      systemId: 'production',
      deviceUuid: '5678',
    });
    const spawner = getSpawner({ fileExists: true, readFileOut });
    const deviceUuid = '1234';

    await runAndCatch(writeOrValidateDeviceCfgFile, { spawner, deviceUuid });

    expect(spawner.exists).toBeCalled();
    expect(spawner.readFile).toBeCalled();
  });

  test('File exists with matching device UUID', async () => {
    const readFileOut = JSON.stringify({
      systemId: 'production',
      deviceUuid: '1234',
    });
    const spawner = getSpawner({ fileExists: true, readFileOut });
    const deviceUuid = '1234';

    await writeOrValidateDeviceCfgFile({ spawner, deviceUuid });

    expect(spawner.exists).toBeCalled();
    expect(spawner.readFile).toBeCalled();
  });
});
