import { join } from 'path';
import { stat, createReadStream } from 'fs';
import fetch from 'node-fetch';
import { CLOUD_API_MODEL_VERSION_PACKAGES_PATH, CloudApiUrl } from '@alwaysai/cloud-api';
import { CliTerseError } from '@alwaysai/alwayscli';
import * as tar from 'tar';
import * as tempy from 'tempy';
import { promisify } from 'util';
import { ModelJsonFile } from './model-json-file';
import { JsSpawner, throwIfNotOk, copyFiles } from '../../util';
import { CliRpcClient, CliAuthenticationClient, getSystemId } from '../../infrastructure';
import { ModelId } from './model-id';

export const modelPackageCloudClient = {
  async download(id: string, version: number) {
    try {
      const { uuid } = await CliRpcClient().getModelVersion({ id, version });
      const modelPackageUrl = ModelPackageUrl(uuid);
      const authorizationHeader =
        await CliAuthenticationClient().getAuthorizationHeader();
      const response = await fetch(modelPackageUrl, {
        method: 'GET',
        headers: { ...authorizationHeader, 'Content-Length': '0' },
      });
      await throwIfNotOk(response);
      return response.body;
    } catch (exception) {
      throw new CliTerseError(`The model ${id} is not associated with your account!`);
    }
  },

  async publish(dir = process.cwd()) {
    const modelJson = ModelJsonFile(dir).read();
    const rpcClient = CliRpcClient();
    const { uuid } = await rpcClient.createModelVersion(modelJson);

    const authorizationHeader = await CliAuthenticationClient().getAuthorizationHeader();

    // Ensure top level directory matches model name
    const { name } = ModelId.parse(modelJson.id);
    const tmpDir = tempy.directory({ prefix: 'alwaysai' });
    const tmpModelDir = join(tmpDir, name);
    await JsSpawner().mkdirp(tmpModelDir);
    await copyFiles(JsSpawner({ path: dir }), JsSpawner({ path: tmpModelDir }));

    const modelPackagePath = tempy.file();
    await tar.create({ cwd: tmpDir, gzip: true, file: modelPackagePath }, [name]);

    await JsSpawner().rimraf(tmpDir);

    const stats = await promisify(stat)(modelPackagePath);

    const modelPackageStream = createReadStream(modelPackagePath);
    const response = await fetch(ModelPackageUrl(uuid), {
      method: 'PUT',
      headers: {
        'Content-Type': 'application/gzip',
        'Content-Length': stats.size.toString(),
        ...authorizationHeader,
      },
      body: modelPackageStream,
    });

    await throwIfNotOk(response);
    await rpcClient.finalizeModelVersion(uuid);
    return uuid;
  },
};

function ModelPackageUrl(uuid: string) {
  const cloudApiUrl = CloudApiUrl(getSystemId());
  return `${cloudApiUrl}${CLOUD_API_MODEL_VERSION_PACKAGES_PATH}/${uuid}`;
}
