import { CliLeaf, CliStringInput, CliUsageError } from '@alwaysai/alwayscli';
import * as chalk from 'chalk';
import { Choice } from 'prompts';
import { yesCliInput } from '../../cli-inputs';
import { checkUserIsLoggedInComponent } from '../../components/user';
import { RequiredWithYesMessage, echo, logger, promptForInput } from '../../util';
import {
  DockerStartParams,
  OpenTunnelResponseBody,
  getDockerImageName,
  getSourceAccessToken,
  processOpenTunnelData,
  removeLocalhostPortsFromKnownHostsFile,
  secureTunnelJsonFile,
  startDockerLocalProxyContainer,
  startSecureTunnelSshSession,
  stopDockerLocalProxyContainer,
} from '../../util/secure-tunnel';
import { getDeviceList } from './list';

// ----------------------------------------------------------------------------
// Local types and interface
// ----------------------------------------------------------------------------
type ConnectedDevice = {
  uuid: string;
  friendlyName: string;
};

// ----------------------------------------------------------------------------
// Local functions
// ----------------------------------------------------------------------------
/**
 * Gets list of connected devices (uuid and device friendly name only)
 * @param {list} deviceList - list connected devices
 * @returns {ConnectedDevice[]} filters out disconnected devices and returns a list of connected devices
 */
function getConnectedDevices(deviceList): ConnectedDevice[] {
  const connectedDevices = deviceList.filter((device) => device.status === 'connected');
  return connectedDevices.map((device) => ({
    uuid: device.uuid,
    friendlyName: device.friendly_name,
  }));
}

/**
 * Prompts user to select the device to connect to
 * @param {ConnectedDevice[]} connectedDevices - list of connected devices
 * @returns {string} uuid of the selected device
 */
async function selectDeviceToConnect(
  connectedDevices: ConnectedDevice[],
): Promise<string> {
  const choices: Choice[] = [{ title: 'None', value: '' }];
  connectedDevices.forEach((device) => {
    choices.push({
      title: `${device.friendlyName}\t(${device.uuid})`,
      value: `${device.uuid}`,
    });
  });

  const choice = await promptForInput({
    purpose: 'to select target device to connect',
    questions: [
      {
        type: 'select',
        name: 'targetDevice',
        message: chalk.yellowBright.bold(
          'Select device to ssh to (disconnected devices are not listed)',
        ),
        initial: 0,
        choices,
      },
    ],
  });

  return choice.targetDevice;
}

/**
 * Prompts user to enter user name for the remote host
 * @returns {string} user name for the remote host
 */
async function getRemoteUserLogin(): Promise<string> {
  const remoteUserLoginAnswer = await promptForInput({
    purpose: 'to set a remote user login',
    questions: [
      {
        type: 'text',
        name: 'login',
        message: chalk.yellowBright.bold('Enter a remote user login'),
      },
    ],
  });
  return remoteUserLoginAnswer.login;
}

// ----------------------------------------------------------------------------
// Main function
// ----------------------------------------------------------------------------
export const deviceConnect = CliLeaf({
  name: 'connect',
  description: 'Connect to remote devices',
  namedInputs: {
    yes: yesCliInput,
    device: CliStringInput({
      description: 'Device UUID',
    }),
    username: CliStringInput({
      description: 'Device username',
    }),
  },
  hidden: true,
  async action(_, { yes, device, username }) {
    const deletedPorts = secureTunnelJsonFile().removeExpiredItems();
    await removeLocalhostPortsFromKnownHostsFile(deletedPorts);

    if (yes) {
      if (device === undefined) {
        throw new CliUsageError(RequiredWithYesMessage('device', undefined));
      }
      if (username === undefined) {
        throw new CliUsageError(RequiredWithYesMessage('username', undefined));
      }
    }

    await checkUserIsLoggedInComponent({ yes });
    let selectedDeviceUuid = device;
    if (selectedDeviceUuid === undefined) {
      const deviceList = await getDeviceList();
      logger.debug(JSON.stringify(deviceList, null, 4));
      if (deviceList.length === 0) {
        return echo(chalk.redBright.bold('No devices found'));
      }

      const connectedDevices = getConnectedDevices(deviceList);
      logger.debug(`connectedDevices = ${JSON.stringify(connectedDevices, null, 4)}`);
      if (connectedDevices.length === 0) {
        return echo(chalk.redBright.bold('No connected devices found'));
      }
      selectedDeviceUuid = await selectDeviceToConnect(connectedDevices);
    }

    logger.debug(`selectedDevice = ${selectedDeviceUuid}`);
    if (selectedDeviceUuid === '') {
      return echo(chalk.redBright.bold('No device selected'));
    }

    const openTunnelData: OpenTunnelResponseBody = await getSourceAccessToken(
      selectedDeviceUuid,
    );
    logger.debug(`openTunnelData = ${JSON.stringify(openTunnelData, null, 4)}`);

    const secureTunnelInfoFromAWS = processOpenTunnelData(openTunnelData);
    if (secureTunnelInfoFromAWS.sourceAccessToken.length > 0) {
      const remoteUserLogin = username ? username : await getRemoteUserLogin();
      logger.debug(`remoteUserLogin = ${remoteUserLogin}`);
      const secureTunnelInfoFromFile = secureTunnelJsonFile().getItem(selectedDeviceUuid);
      secureTunnelInfoFromAWS.sshPort =
        Object.keys(secureTunnelInfoFromFile).length > 0
          ? secureTunnelInfoFromFile.sshPort
          : secureTunnelJsonFile().getNextAvailablePort();
      secureTunnelJsonFile().setItem(selectedDeviceUuid, secureTunnelInfoFromAWS);

      const startParams: DockerStartParams = {
        deviceUuid: selectedDeviceUuid,
        imageName: getDockerImageName(),
        sshPort: secureTunnelInfoFromAWS.sshPort,
        sourceAccessToken: secureTunnelInfoFromAWS.sourceAccessToken,
      };
      await startDockerLocalProxyContainer(startParams);
      await startSecureTunnelSshSession(remoteUserLogin, secureTunnelInfoFromAWS.sshPort);
      await removeLocalhostPortsFromKnownHostsFile([secureTunnelInfoFromAWS.sshPort]);
      await stopDockerLocalProxyContainer(selectedDeviceUuid);
    }
  },
});
