import * as fs from 'fs';
import { testSuiteFuncs } from './validate-package';
import { CLI_TERSE_ERROR } from '@alwaysai/alwayscli';
import { runAndCatch } from '@carnesen/run-and-catch';

jest.mock('fs');
const mockFs = jest.mocked(fs, true);

const mockModelJsonFileRead = jest.fn().mockReturnValue({});
const mockModelJsonFileExists = jest.fn().mockReturnValue(true);
const mockModelJsonGetErrors = jest.fn().mockReturnValue([]);

jest.mock('../../core/model/model-package-json-file', () => {
  return {
    ModelPackageJsonFile: jest.fn().mockImplementation(() => {
      return {
        read: mockModelJsonFileRead,
        exists: mockModelJsonFileExists,
        getErrors: mockModelJsonGetErrors,
      };
    }),
  };
});

const correctJson = {
  id: 'alwaysai/mobilenet_ssd',
  accuracy: '',
  description:
    'Classifies 20 PASCAL object classes: person, bird, cat, cow, dog, horse, sheep, aeroplane, bicycle, boat, bus, car, motorbike, train, bottle, chair, dining table, potted plant, sofa, tv/monitor.',
  inference_time: 0.18125,
  license: 'MIT',
  mean_average_precision_top_1: 0.727,
  mean_average_precision_top_5: null,
  public: true,
  dataset: 'Pascal VOC',
  website_url: 'https://github.com/chuanqi305/MobileNet-SSD',
  model_parameters: {
    framework_type: 'caffe',
    config_file: 'models/MobileNetSSD_deploy.prototxt.txt',
    size: [300, 300],
    model_file: 'models/MobileNetSSD_deploy.caffemodel',
    label_file: 'labels/pascal_voc.txt',
    scalefactor: 0.007843,
    mean: [127.5, 127.5, 127.5],
    crop: true,
    swaprb: false,
    purpose: 'ObjectDetection',
  },
  version: 4,
};

describe(__filename, () => {
  afterEach(() => {
    jest.clearAllMocks();
  });

  beforeEach(() => {
    mockFs.existsSync.mockReturnValue(true);

    mockModelJsonFileExists.mockReturnValue(true);
    mockModelJsonFileRead.mockReturnValue(correctJson);
  });

  it('tests a missing model configuration', async () => {
    mockModelJsonFileExists.mockReturnValue(false);
    const ex = await runAndCatch(testSuiteFuncs.validatePackageComponent);
    expect(ex.code).toBe(CLI_TERSE_ERROR);
    expect(mockFs.existsSync).toBeCalledTimes(0);
    expect(mockModelJsonGetErrors).toBeCalledTimes(0);
  });

  it('tests a valid model configuration', async () => {
    expect(await testSuiteFuncs.validatePackageComponent()).toBeTruthy();
    expect(mockFs.existsSync).toBeCalled();
    expect(mockModelJsonGetErrors).toBeCalledTimes(0);
  });

  it('tests an invalid model configuration', async () => {
    mockModelJsonFileRead.mockImplementation(() => {
      throw new Error();
    });
    expect(await testSuiteFuncs.validatePackageComponent()).toBeFalsy();
    expect(mockFs.existsSync).toBeCalledTimes(0);
    expect(mockModelJsonGetErrors).toBeCalled();
  });

  it('tests an invalid model ID', async () => {
    const incorrectJson = JSON.parse(JSON.stringify(correctJson));
    incorrectJson['id'] = 'invalid/id/here';
    mockModelJsonFileRead.mockReturnValue(incorrectJson);
    expect(await testSuiteFuncs.validatePackageComponent()).toBeFalsy();
    expect(mockFs.existsSync).toBeCalled();
    expect(mockModelJsonGetErrors).toBeCalledTimes(0);
  });

  it('tests a missing file', async () => {
    // mock the fs to return false, the filepath is not correct.
    mockFs.existsSync.mockReturnValue(false);
    expect(await testSuiteFuncs.validatePackageComponent()).toBeFalsy();
    expect(mockFs.existsSync).toBeCalled();
    expect(mockModelJsonGetErrors).toBeCalledTimes(0);
  });

  it('tests filepath set to "null"', async () => {
    // Make a copy of incorrect JSON and set value of config file to null
    const incorrectJson = JSON.parse(JSON.stringify(correctJson));
    incorrectJson['model_parameters']['label_file'] = null;
    expect(await testSuiteFuncs.validatePackageComponent()).toBeTruthy();
    expect(mockFs.existsSync).toBeCalled();
    expect(mockModelJsonGetErrors).toBeCalledTimes(0);
  });
});
