import fetch, { FetchError, Response } from 'node-fetch';
import { promisify } from 'util';
import { dirname } from 'path';
import { createWriteStream, rename } from 'fs';
import { CodedError } from '@carnesen/coded-error';
import mkdirp = require('mkdirp');
import pump = require('pump');
import rimraf from 'rimraf';
import { CliTerseError } from '@alwaysai/alwayscli';

import { RandomString } from './random-string';
import { PLEASE_REPORT_THIS_ERROR_MESSAGE } from '../constants';

async function download(url: string) {
  const response = await fetch(url, {
    method: 'GET',
  });
  await throwIfNotOk(response);
  return response.body;
}

export async function throwIfNotOk(response: Response) {
  if (!response.ok) {
    const parsed = (await response.json()) as FetchError;
    const {
      message = `Server responded ${response.status} ("${response.statusText}")`,
      code,
    } = parsed;
    throw new CodedError(message, code);
  }
}

async function writeToFile(path: string, readable: NodeJS.ReadableStream) {
  mkdirp.sync(dirname(path));
  const tmpFilePath = `${path}.${RandomString()}.tmp`;
  try {
    await new Promise<void>((resolve, reject) => {
      const writeable = createWriteStream(tmpFilePath);
      pump(readable, writeable, (err) => {
        if (err) {
          reject(err);
        } else {
          resolve();
        }
      });
    });
    await promisify(rename)(tmpFilePath, path);
  } catch (exception) {
    try {
      await rimraf(tmpFilePath);
    } finally {
      throw exception;
    }
  }
}

export async function downloadToFile(url: string, writePath: string) {
  const fileContents = await download(url);
  if (fileContents === null) {
    throw new CliTerseError(
      `Error downloading ${url}. ${PLEASE_REPORT_THIS_ERROR_MESSAGE}`,
    );
  }
  await writeToFile(writePath, fileContents);
}
