import { SecureTunnelInteractiveSsh } from './secure-tunnel-ssh';
import { PRIVATE_KEY_FILE_PATH } from '../paths';
import { spawn, ChildProcess } from 'child_process';
// import { mocked } from 'ts-jest/utils';
import { PassThrough } from 'stream';

// ----------------------------------------------------------------------------
// Local constants
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// Mocks for the SecureTunnelInteractiveSsh
// ----------------------------------------------------------------------------
jest.mock('child_process', () => ({
  spawn: jest.fn(),
}));
jest.mock('../paths', () => ({
  PRIVATE_KEY_FILE_PATH: '/unit/test/path/to/private/key',
}));

// ----------------------------------------------------------------------------
// Tests for SecureTunnelInteractiveSsh
// ----------------------------------------------------------------------------
describe('SecureTunnelInteractiveSsh', () => {
  const testUserName = 'testUser';
  const testSshPort = 5042;

  describe('processArgs', () => {
    beforeEach(() => {
      jest.clearAllMocks();
    });

    it('should process arguments correctly', () => {
      const sshInstance = new SecureTunnelInteractiveSsh({
        targetHost: `${testUserName}@localhost`,
        sshPort: testSshPort,
      });

      const processedArgs = sshInstance.processArgs([]);

      expect(processedArgs).toEqual([
        '-i',
        PRIVATE_KEY_FILE_PATH,
        '-o',
        'StrictHostKeyChecking=no',
        '-t',
        '-p',
        `${testSshPort}`,
        `${testUserName}@localhost`,
      ]);
    });

    it('should process additional arguments correctly', () => {
      const sshInstance = new SecureTunnelInteractiveSsh({
        targetHost: `${testUserName}@localhost`,
        sshPort: testSshPort,
      });
      const additionalArgs = ['-just_a_test', '-alwaysai'];

      const processedArgs = sshInstance.processArgs(additionalArgs);

      expect(processedArgs).toEqual([
        '-i',
        PRIVATE_KEY_FILE_PATH,
        '-o',
        'StrictHostKeyChecking=no',
        '-t',
        '-p',
        `${testSshPort}`,
        ...additionalArgs,
        `${testUserName}@localhost`,
      ]);
    });
  });

  describe('runInteractiveSshAsync', () => {
    let mockSshSessionOn;

    beforeEach(() => {
      jest.clearAllMocks();
    });

    it('should run the ssh session', async () => {
      const sshInstance = new SecureTunnelInteractiveSsh({
        targetHost: `${testUserName}@localhost`,
        sshPort: testSshPort,
      });
      mockSshSessionOn = jest.fn();
      (spawn as jest.Mock).mockImplementationOnce(() => ({
        on: mockSshSessionOn,
      }));
      const mockChildProcess: Partial<ChildProcess> = {
        on: jest.fn(),
        stdout: new PassThrough(),
        stderr: new PassThrough(),
      };
      const testSshArgs = ['arg1', 'arg2'];

      const mockSpawn = spawn as jest.MockedFunction<typeof spawn>;
      mockSpawn.mockReturnValueOnce(mockChildProcess as ChildProcess);

      sshInstance.runInteractiveSshAsync(testSshArgs);

      expect(spawn).toBeCalledTimes(1);
      expect(spawn).toHaveBeenCalledWith(`ssh ${testSshArgs.join(' ')}`, {
        stdio: [process.stdin, process.stdout, process.stderr],
        shell: true,
      });
      expect(mockSshSessionOn).toHaveBeenCalledTimes(2);
      expect(mockSshSessionOn).toHaveBeenCalledWith('exit', expect.any(Function));
      expect(mockSshSessionOn).toHaveBeenCalledWith('close', expect.any(Function));
    });
  });
});
