import {
  getDockerImageName,
  removeLocalhostPortsFromKnownHostsFile,
  secureTunnelJsonFile,
  getSourceAccessToken,
  OpenTunnelResponseBody,
  startDockerLocalProxyContainer,
  stopDockerLocalProxyContainer,
  startSecureTunnelSshSession,
  processOpenTunnelData,
} from './secure-tunnel';
import { ConfigFileSchema } from '@alwaysai/config-nodejs';
import fetch from 'node-fetch';
import { serviceEndpointBuilder } from '../infrastructure/urls';
import { CliAuthenticationClient } from '../infrastructure/authentication-client';
import { v4 as uuidv4 } from 'uuid';
import { Spinner, echo } from './';
import { delay } from './delay';
import { JsSpawner } from './spawner';
import * as boxen from 'boxen';
import * as chalk from 'chalk';
import { CliTerseError } from '@alwaysai/alwayscli';
import { SecureTunnelInteractiveSsh } from './secure-tunnel-ssh';

// ----------------------------------------------------------------------------
// Local constants
// ----------------------------------------------------------------------------
const TEST_ST_DEVICE_UUID = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa';
const TEST_ST_CONFIG_FILE_OBJ = {
  expiresAt: Math.floor(Date.now() / 1000) + 3600,
  sshPort: 8001,
  sourceAccessToken: 'test_sourceAccessToken',
  tunnelId: 'dddddddd-dddd-dddd-dddd-dddddddddddd',
};

// ----------------------------------------------------------------------------
// Mocks for the ConfigFileSchema
// ----------------------------------------------------------------------------
jest.mock('@alwaysai/config-nodejs');

// ----------------------------------------------------------------------------
// Tests for secureTunnelJsonFile
// ----------------------------------------------------------------------------
describe('secureTunnelJsonFile', () => {
  let mockInitialize;
  let mockUpdate;
  let mockRead;

  beforeEach(() => {
    jest.clearAllMocks();
    mockInitialize = jest.fn();
    mockUpdate = jest.fn();
    mockRead = jest.fn();
    (ConfigFileSchema as jest.Mock).mockReturnValue({
      initialize: mockInitialize,
      update: mockUpdate,
      read: mockRead,
    });
  });

  it('should set an item in the secure tunnel storage', () => {
    secureTunnelJsonFile().setItem(TEST_ST_DEVICE_UUID, TEST_ST_CONFIG_FILE_OBJ);

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledWith(expect.any(Function));
    expect(mockRead).toHaveBeenCalledTimes(0);
  });

  it('should get an item from the secure tunnel storage', () => {
    mockRead.mockReturnValueOnce({ [TEST_ST_DEVICE_UUID]: TEST_ST_CONFIG_FILE_OBJ });

    const secureTunnelItem = secureTunnelJsonFile().getItem(TEST_ST_DEVICE_UUID);

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(0);
    expect(mockRead).toHaveBeenCalledTimes(1);
    expect(secureTunnelItem).toEqual(TEST_ST_CONFIG_FILE_OBJ);
  });

  it('should remove an item from the secure tunnel storage', () => {
    secureTunnelJsonFile().removeItem(TEST_ST_DEVICE_UUID);

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledWith(expect.any(Function));
    expect(mockRead).toHaveBeenCalledTimes(0);
  });

  it('should remove expired items from the secure tunnel storage', () => {
    const TEST_ST_DEVICE_UUID_TO_REMOVE_1 = 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb';
    const TEST_EXPIRED_TIME_STAMP = Math.floor(Date.now() / 1000) - 3600;
    const TEST_ST_CONFIG_FILE_OBJ_TO_REMOVE_1 = {
      expiresAt: TEST_EXPIRED_TIME_STAMP,
      sshPort: 8002,
      sourceAccessToken: 'test_sourceAccessToken_to_remove_1',
      tunnelId: 'eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee',
    };
    const TEST_ST_DEVICE_UUID_TO_REMOVE_2 = 'cccccccc-cccc-cccc-cccc-cccccccccccc';
    const TEST_ST_CONFIG_FILE_OBJ_TO_REMOVE_2 = {
      expiresAt: TEST_EXPIRED_TIME_STAMP,
      sshPort: 8003,
      sourceAccessToken: 'test_sourceAccessToken_to_remove_2',
      tunnelId: 'ffffffff-ffff-ffff-ffff-ffffffffffff',
    };
    const testData = {
      [TEST_ST_DEVICE_UUID_TO_REMOVE_1]: TEST_ST_CONFIG_FILE_OBJ_TO_REMOVE_1,
      [TEST_ST_DEVICE_UUID]: TEST_ST_CONFIG_FILE_OBJ,
      [TEST_ST_DEVICE_UUID_TO_REMOVE_2]: TEST_ST_CONFIG_FILE_OBJ_TO_REMOVE_2,
    };
    mockRead.mockReturnValueOnce(testData);
    let capturedCallback;
    mockUpdate.mockImplementation((callback) => {
      capturedCallback = callback;
    });

    const removedItems = secureTunnelJsonFile().removeExpiredItems();
    capturedCallback(testData);

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(1);
    expect(mockRead).toHaveBeenCalledTimes(0);
    expect(removedItems).toEqual([
      TEST_ST_CONFIG_FILE_OBJ_TO_REMOVE_1.sshPort,
      TEST_ST_CONFIG_FILE_OBJ_TO_REMOVE_2.sshPort,
    ]);
  });

  it('should NOT remove items, if no expired items found in the secure tunnel storage', () => {
    const TEST_ST_DEVICE_UUID_NOT_EXPIRED = 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb';
    const TEST_EXPIRED_TIME_STAMP = Math.floor(Date.now() / 1000) + 1800;
    const TEST_ST_CONFIG_FILE_OBJ_NOT_EXPIRED = {
      expiresAt: TEST_EXPIRED_TIME_STAMP,
      sshPort: 8002,
      sourceAccessToken: 'test_sourceAccessToken_to_remove_1',
      tunnelId: 'eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee',
    };
    const testData = {
      [TEST_ST_DEVICE_UUID_NOT_EXPIRED]: TEST_ST_CONFIG_FILE_OBJ_NOT_EXPIRED,
      [TEST_ST_DEVICE_UUID]: TEST_ST_CONFIG_FILE_OBJ,
    };
    mockRead.mockReturnValueOnce(testData);
    let capturedCallback;
    mockUpdate.mockImplementation((callback) => {
      capturedCallback = callback;
    });

    const removedItems = secureTunnelJsonFile().removeExpiredItems();
    capturedCallback(testData);

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(1);
    expect(mockRead).toHaveBeenCalledTimes(0);
    expect(removedItems).toEqual([]);
  });

  it('should get the next available port', () => {
    const TEST_ST_DEVICE_UUID_FOR_MAX_PORT = 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb';
    const TEST_EXPIRED_TIME_STAMP = Math.floor(Date.now() / 1000) + 1800;
    const TEST_ST_CONFIG_FILE_OBJ_FOR_MAX_PORT = {
      expiresAt: TEST_EXPIRED_TIME_STAMP,
      sshPort: 8005,
      sourceAccessToken: 'test_sourceAccessToken_for_max_port',
      tunnelId: 'eeeeeeee-eeee-eeee-eeee-eeeeeeeeeeee',
    };
    const testData = {
      [TEST_ST_DEVICE_UUID_FOR_MAX_PORT]: TEST_ST_CONFIG_FILE_OBJ_FOR_MAX_PORT,
      [TEST_ST_DEVICE_UUID]: TEST_ST_CONFIG_FILE_OBJ,
    };
    const expectedNextAvailablePort =
      Math.max(...Object.values(testData).map((item) => item.sshPort)) + 1;
    mockRead.mockReturnValueOnce(testData);

    const actualNextAvailablePort = secureTunnelJsonFile().getNextAvailablePort();

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(0);
    expect(mockRead).toHaveBeenCalledTimes(1);
    expect(actualNextAvailablePort).toEqual(expectedNextAvailablePort);
  });

  it('should get the initial next available port, if no previous data found', () => {
    const testData = {};
    const expectedNextAvailablePort = 5001;
    mockRead.mockReturnValueOnce(testData);

    const actualNextAvailablePort = secureTunnelJsonFile().getNextAvailablePort();

    expect(mockInitialize).toHaveBeenCalledTimes(1);
    expect(mockUpdate).toHaveBeenCalledTimes(0);
    expect(mockRead).toHaveBeenCalledTimes(1);
    expect(actualNextAvailablePort).toEqual(expectedNextAvailablePort);
  });

  it('should handle ENOENT error', () => {
    mockInitialize.mockImplementation(() => {
      throw new Error('ENOENT');
    });
    let errorCatch;

    try {
      secureTunnelJsonFile();
      expect(true).toBe(false); // should not reach this line
    } catch (e) {
      errorCatch = e;
      expect(errorCatch.message).toEqual('ENOENT');
    }
  });

  it('should handle EACCES error', () => {
    mockInitialize.mockImplementation(() => {
      throw new Error('EACCES');
    });
    let errorCatch;

    try {
      secureTunnelJsonFile();
      expect(true).toBe(false); // should not reach this line
    } catch (e) {
      errorCatch = e;
      expect(errorCatch.message).toEqual('EACCES');
    }
  });
});

// ----------------------------------------------------------------------------
// Mocks for getDockerImageName
// ----------------------------------------------------------------------------
jest.mock('../infrastructure', () => {
  return {
    ...jest.requireActual('../infrastructure'),
    getSystemId: jest.fn(),
  };
});

// ----------------------------------------------------------------------------
// Tests for getDockerImageName
// ----------------------------------------------------------------------------
describe('getDockerImageName', () => {
  const testCases: [string, string][] = [
    // env,           ext
    ['development', '-dev'],
    ['qa', '-qa'],
    ['production', ''],
    ['', ''],
    ['anything-else', ''],
  ];

  beforeEach(() => {
    jest.clearAllMocks();
  });

  test.each(testCases)(
    'should return the correct Docker image name for %s',
    (environment, expectedExt) => {
      const { getSystemId } = require('../infrastructure');
      getSystemId.mockReturnValue(environment);
      const expectedDockerImageName = `alwaysai/securetunnel-localproxy${expectedExt}:ubuntu-20.04-latest`;

      const actualResult = getDockerImageName();

      expect(actualResult).toEqual(expectedDockerImageName);
    },
  );
});

// ----------------------------------------------------------------------------
// Mocks for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
jest.mock('./spawner');
const mockJsSpawner = {
  exists: jest.fn(),
  readFile: jest.fn(),
  writeFile: jest.fn(),
  run: jest.fn(),
};
jest.mock('./spawner', () => {
  return {
    ...jest.requireActual('./spawner'),
    JsSpawner: jest.fn(() => mockJsSpawner),
  };
});

const pathToKnownHostsFile = '/path/to/known_hosts';
jest.mock('../paths', () => {
  return {
    ...jest.requireActual('../paths'),
    DOT_SSH_DIR: '/path/to/known_hosts',
  };
});

// ----------------------------------------------------------------------------
// Tests for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
describe('removeLocalhostPortsFromKnownHostsFile', () => {
  beforeEach(() => {
    jest.clearAllMocks();
  });

  it('should remove localhost ports from known_hosts file', async () => {
    const sshPortListToRemove = [8001, 8042];
    const hostNotToDelete =
      'example.com ssh-rsa fingerPrintForSomeOtherHost_NotToDelete\n';
    const knownHostsContent = [
      `[localhost]:${sshPortListToRemove[0]} ssh-rsa fingerPrintForLocalHostPort${sshPortListToRemove[0]}\n`,
      hostNotToDelete,
      `[localhost]:${sshPortListToRemove[1]} ssh-rsa fingerPrintForLocalHostPort${sshPortListToRemove[1]}\n`,
    ];
    mockJsSpawner.exists.mockResolvedValue(true);
    mockJsSpawner.readFile.mockResolvedValue(knownHostsContent.join(''));

    await removeLocalhostPortsFromKnownHostsFile(sshPortListToRemove);

    expect(mockJsSpawner.exists).toHaveBeenCalledWith(
      `${pathToKnownHostsFile}/known_hosts`,
    );
    expect(mockJsSpawner.readFile).toHaveBeenCalledWith(
      `${pathToKnownHostsFile}/known_hosts`,
    );
    expect(mockJsSpawner.writeFile).toHaveBeenCalledWith(
      `${pathToKnownHostsFile}/known_hosts`,
      hostNotToDelete,
    );
  });

  it('should do nothing if known_hosts file does not exist', async () => {
    mockJsSpawner.exists.mockResolvedValue(false);
    const sshPortListToRemove = [8001, 8042];

    await removeLocalhostPortsFromKnownHostsFile(sshPortListToRemove);

    expect(mockJsSpawner.exists).toHaveBeenCalledWith(
      `${pathToKnownHostsFile}/known_hosts`,
    );
    expect(mockJsSpawner.readFile).not.toHaveBeenCalled();
    expect(mockJsSpawner.writeFile).not.toHaveBeenCalled();
  });

  it('should do nothing, given ports to remove list is empty', async () => {
    const sshPortListToRemove = [];
    const hostNotToDelete =
      'example.com ssh-rsa fingerPrintForSomeOtherHost_NotToDelete\n';
    const knownHostsContent = [hostNotToDelete];
    mockJsSpawner.exists.mockResolvedValue(true);
    mockJsSpawner.readFile.mockResolvedValue(knownHostsContent.join(''));

    await removeLocalhostPortsFromKnownHostsFile(sshPortListToRemove);

    expect(mockJsSpawner.exists).not.toHaveBeenCalledWith(
      `${pathToKnownHostsFile}/known_hosts`,
    );
    expect(mockJsSpawner.readFile).not.toHaveBeenCalled();
    expect(mockJsSpawner.writeFile).not.toHaveBeenCalled();
  });
});

// ----------------------------------------------------------------------------
// Mocks for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
jest.mock('node-fetch');
jest.mock('../infrastructure/authentication-client');
jest.mock('../infrastructure/urls');
jest.mock('uuid');

// ----------------------------------------------------------------------------
// Tests for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
describe('getSourceAccessToken', () => {
  const expectedDeviceUuid = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa';
  const expectedAccessToken: OpenTunnelResponseBody = {
    expiresAt: Math.floor(Date.now() / 1000) + 12 * 3600,
    newTunnelCreated: true,
    sourceAccessToken: 'your-mocked-source-access-token',
    txId: 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb',
    tunnelId: 'cccccccc-cccc-cccc-cccc-cccccccccccc',
  };
  const expectedAuthorizationHeader = 'test-authorization-header';
  const mockGetIdAuthorizationHeader = jest.fn().mockResolvedValueOnce({
    Authorization: expectedAuthorizationHeader,
  });
  (CliAuthenticationClient as jest.Mock).mockImplementation(() => ({
    getIdAuthorizationHeader: mockGetIdAuthorizationHeader,
  }));
  const expectedUrl = 'test-mocked-request-url';
  (serviceEndpointBuilder as jest.Mock).mockReturnValue(expectedUrl);
  const expectedUuid = 'your-mocked-uuid';

  beforeEach(() => {
    jest.clearAllMocks();
  });

  it('should fetch and parse the response correctly', async () => {
    (uuidv4 as jest.Mock).mockReturnValue(expectedUuid);
    const mockResponse = {
      status: 200,
      buffer: jest
        .fn()
        .mockResolvedValueOnce(Buffer.from(JSON.stringify(expectedAccessToken))),
    };
    (fetch as any as jest.Mock).mockResolvedValueOnce(mockResponse);

    const result = await getSourceAccessToken(expectedDeviceUuid);

    expect(result).toEqual(expectedAccessToken);
    expect(mockGetIdAuthorizationHeader).toHaveBeenCalled();
    expect(serviceEndpointBuilder).toHaveBeenCalledWith(
      'secure-tunnel',
      'openSecureTunnel',
    );
    expect(fetch).toHaveBeenCalledWith(expectedUrl, {
      method: 'post',
      body: JSON.stringify({ deviceUuid: expectedDeviceUuid, txId: expectedUuid }),
      headers: {
        Authorization: expectedAuthorizationHeader,
        'Content-Type': 'application/json',
      },
    });
  });

  it('should handle an invalid response', async () => {
    (fetch as any as jest.Mock).mockResolvedValueOnce(null);

    try {
      await getSourceAccessToken('your-device-uuid');
      expect(true).toBe(false); // Should not reach here
    } catch (error) {
      expect(error.message).toBe('Error: Receiving response from server');
    }

    expect(mockGetIdAuthorizationHeader).toHaveBeenCalled();
    expect(serviceEndpointBuilder).toHaveBeenCalledWith(
      'secure-tunnel',
      'openSecureTunnel',
    );
  });

  it('should handle error http status forbidden response', async () => {
    const mockResponse = {
      status: 403,
      buffer: jest.fn().mockResolvedValueOnce(null),
    };
    (fetch as any as jest.Mock).mockResolvedValueOnce(mockResponse);

    try {
      await getSourceAccessToken('your-device-uuid');
      expect(true).toBe(false); // Should not reach here
    } catch (error) {
      expect(error.message).toBe('Error: Unauthorized, response.status = 403');
    }

    expect(mockGetIdAuthorizationHeader).toHaveBeenCalled();
    expect(serviceEndpointBuilder).toHaveBeenCalledWith(
      'secure-tunnel',
      'openSecureTunnel',
    );
  });
});

// ----------------------------------------------------------------------------
// Mocks for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
jest.mock('./spinner');
jest.mock('./delay', () => ({
  delay: jest.fn(() => Promise.resolve()),
}));
jest.mock('./echo', () => ({
  echo: jest.fn(),
}));

// ----------------------------------------------------------------------------
// Tests for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
describe('startDockerLocalProxyContainer', () => {
  const startParams = {
    deviceUuid: 'test-start-docker-device-uuid',
    sshPort: 8042,
    imageName: 'test-start-docker-image-name',
    sourceAccessToken: 'test-start-docker-source-access-token',
  };
  let mockSucceed;
  let mockFail;

  beforeEach(() => {
    jest.clearAllMocks();
    mockSucceed = jest.fn();
    mockFail = jest.fn();
    (Spinner as jest.Mock).mockReturnValue({
      succeed: mockSucceed,
      fail: mockFail,
    });
  });

  it('should start docker container and succeed', async () => {
    console.log('should start a Docker container and succeed');
    (JsSpawner().run as jest.Mock).mockResolvedValueOnce(undefined);

    await startDockerLocalProxyContainer(startParams);

    expect(Spinner).toHaveBeenCalledWith('Start docker container');
    expect(JsSpawner().run).toHaveBeenCalledWith({
      exe: 'docker',
      args: [
        'run',
        '--rm',
        '--detach',
        '--name',
        startParams.deviceUuid,
        '--expose',
        `${startParams.sshPort}`,
        '-p',
        `${startParams.sshPort}:${startParams.sshPort}`,
        startParams.imageName,
        '-s',
        `${startParams.sshPort}`,
        '-t',
        startParams.sourceAccessToken,
      ],
    });
    expect(delay).toHaveBeenCalledWith(1000);
    expect(mockSucceed).toHaveBeenCalled();
    expect(mockFail).not.toHaveBeenCalled();
    expect(echo).not.toHaveBeenCalled();
  });

  it('start docker should handle an error and fail', async () => {
    const errorMessage = 'Docker run failed';
    (JsSpawner().run as jest.Mock).mockRejectedValueOnce(new Error(errorMessage));

    try {
      await startDockerLocalProxyContainer(startParams);
      expect(true).toBe(false); // should not reach this line
    } catch (error) {
      expect(Spinner).toHaveBeenCalledWith('Start docker container');
      expect(JsSpawner().run).toHaveBeenCalledWith({
        exe: 'docker',
        args: [
          'run',
          '--rm',
          '--detach',
          '--name',
          startParams.deviceUuid,
          '--expose',
          `${startParams.sshPort}`,
          '-p',
          `${startParams.sshPort}:${startParams.sshPort}`,
          startParams.imageName,
          '-s',
          `${startParams.sshPort}`,
          '-t',
          startParams.sourceAccessToken,
        ],
      });
      expect(delay).not.toHaveBeenCalled();
      expect(mockSucceed).not.toHaveBeenCalled();
      expect(mockFail).toHaveBeenCalledWith(
        `Command "docker run ${startParams.imageName}" failed`,
      );
      expect(echo).toBeCalledTimes(6);
      expect(echo).toHaveBeenCalledWith(
        chalk.redBright.bold(`Failed to start docker container: ${errorMessage}`),
      );
      expect(echo).toHaveBeenCalledWith();
      expect(echo).toHaveBeenCalledWith(
        'Please make sure you installed Docker and Docker engine is running',
      );
      expect(echo).toHaveBeenCalledWith(
        'On Linux: Please double-check that you\'ve completed the "Manage Docker as a non-root user" post-install steps described here:',
      );
      expect(echo).toHaveBeenCalledWith(
        '  https://docs.docker.com/install/linux/linux-postinstall/',
      );
      expect(echo).toHaveBeenCalledWith();
    }
  });
});

// ----------------------------------------------------------------------------
// Tests for removeLocalhostPortsFromKnownHostsFile
// ----------------------------------------------------------------------------
describe('stopDockerLocalProxyContainer', () => {
  const testContainerName = 'test-stop-docker-container-name';
  let mockSucceed;
  let mockFail;

  beforeEach(() => {
    jest.clearAllMocks();
    mockSucceed = jest.fn();
    mockFail = jest.fn();
    (Spinner as jest.Mock).mockReturnValue({
      succeed: mockSucceed,
      fail: mockFail,
    });
  });

  it('should stop docker container and succeed', async () => {
    (JsSpawner().run as jest.Mock).mockResolvedValueOnce(undefined);

    await stopDockerLocalProxyContainer(testContainerName);

    expect(Spinner).toHaveBeenCalledWith(`Stop docker container: ${testContainerName}`);
    expect(JsSpawner().run).toHaveBeenCalledWith({
      exe: 'docker',
      args: ['stop', testContainerName],
    });
    expect(mockSucceed).toHaveBeenCalled();
    expect(mockFail).not.toHaveBeenCalled();
    expect(echo).not.toHaveBeenCalled();
  });

  it('stop docker should handle an error and throw CliTerseError', async () => {
    const errorMessage = 'Docker stop failed';
    (JsSpawner().run as jest.Mock).mockRejectedValueOnce(new Error(errorMessage));

    try {
      await stopDockerLocalProxyContainer(testContainerName);
      expect(true).toBe(false); // should not reach this line
    } catch (error) {
      expect(Spinner).toHaveBeenCalledWith(`Stop docker container: ${testContainerName}`);
      expect(JsSpawner().run).toHaveBeenCalledWith({
        exe: 'docker',
        args: ['stop', testContainerName],
      });
      expect(mockSucceed).not.toHaveBeenCalled();
      expect(mockFail).toHaveBeenCalledWith(
        `Command "docker stop ${testContainerName}" failed`,
      );
      expect(echo).toBeCalledTimes(2);
      expect(echo).toHaveBeenCalledWith(
        chalk.redBright.bold(`Failed to stop docker container: ${errorMessage}`),
      );
      expect(echo).toHaveBeenCalledWith();
      expect(error).toBeInstanceOf(CliTerseError);
      expect(error.message).toBe(`Failed to stop docker container: ${testContainerName}`);
    }
  });
});

// ----------------------------------------------------------------------------
// Mocks for startSecureTunnelSshSession
// ----------------------------------------------------------------------------
jest.mock('./secure-tunnel-ssh');

// ----------------------------------------------------------------------------
// Tests for startSecureTunnelSshSession
// ----------------------------------------------------------------------------
describe('startSecureTunnelSshSession', () => {
  const testUser = 'testUser_startSecureTunnelSshSession';
  const testPort = 5042;
  const testTargetHost = `${testUser}@localhost`;
  const testArgs = ['-i', '/path/to/private/key', '-p', `${testPort}`, testTargetHost];
  let mockProcessArgs;
  let mockRunInteractiveSshAsync;

  beforeEach(() => {
    jest.clearAllMocks();
    mockProcessArgs = jest.fn();
    mockRunInteractiveSshAsync = jest.fn();
    (SecureTunnelInteractiveSsh as jest.Mock).mockImplementationOnce(() => ({
      processArgs: mockProcessArgs,
      runInteractiveSshAsync: mockRunInteractiveSshAsync,
    }));
  });

  it('should start an SSH session successfully', async () => {
    mockProcessArgs.mockReturnValueOnce(testArgs);

    await startSecureTunnelSshSession(testUser, testPort);

    expect(echo).toHaveBeenCalledWith(
      'Please enter the ssh password for remote host if prompted.',
    );
    expect(SecureTunnelInteractiveSsh).toHaveBeenCalledWith({
      targetHost: testTargetHost,
      sshPort: testPort,
    });
    expect(mockProcessArgs).toHaveBeenCalledTimes(1);
    expect(mockProcessArgs).toHaveBeenCalledWith([]);
    expect(mockRunInteractiveSshAsync).toHaveBeenCalledTimes(1);
    expect(mockRunInteractiveSshAsync).toHaveBeenCalledWith(testArgs);
  });

  it('should handle error when starting SSH session', async () => {
    const errorMessage = 'Failed to start SSH session';
    mockProcessArgs.mockReturnValueOnce(testArgs);
    mockRunInteractiveSshAsync.mockRejectedValueOnce(new Error(errorMessage));

    try {
      await startSecureTunnelSshSession(testUser, testPort);
      expect(true).toBe(false); // should not reach this line
    } catch (error) {
      expect(SecureTunnelInteractiveSsh).toHaveBeenCalledWith({
        targetHost: testTargetHost,
        sshPort: testPort,
      });
      expect(mockProcessArgs).toHaveBeenCalledTimes(1);
      expect(mockProcessArgs).toHaveBeenCalledWith([]);
      expect(mockRunInteractiveSshAsync).toHaveBeenCalledTimes(1);
      expect(mockRunInteractiveSshAsync).toHaveBeenCalledWith(testArgs);
      expect(echo).toHaveBeenCalledWith(
        chalk.redBright.bold(`Failed to start ssh session: ${errorMessage}`),
      );
      expect(echo).toHaveBeenCalledWith();
    }
  });
});

// ----------------------------------------------------------------------------
// Tests for processOpenTunnelData
// ----------------------------------------------------------------------------
describe('processOpenTunnelData', () => {
  const expectedCommonMessageToDisplay = `Please make sure the docker engine is running before entering your remote user login.
Please complete your ssh remote session with \"exit\", so that resource can be cleaned up.
Happy Secure Tunneling!\n
In case the ssh session does not start within 10 seconds, it is safe to assume that
target host is not capable to establish a secure tunnel connection. The reason could be
an outdated version of AlwaysAI device-agent on target or target OS/architecture is not supported.
In that case, please quit CLI with Ctrl+C and check the documentation for more details.`;
  const sinceEpochTime = Math.floor(Date.now() / 1000) - 30 * 60;
  const testCasesForUsedBy: [string, number, string][] = [
    // by,          since,          expectedUsedByMessage
    ['', 0, 'Used by: unknown'],
    [
      '',
      sinceEpochTime,
      `Used by: unknown, since: ${new Date(sinceEpochTime * 1000).toLocaleString()}`,
    ],
    ['testUser_3', 0, 'Used by: testUser_3'],
    [
      'testUser_4',
      sinceEpochTime,
      `Used by: testUser_4, since: ${new Date(sinceEpochTime * 1000).toLocaleString()}`,
    ],
  ];

  beforeEach(() => {
    jest.clearAllMocks();
  });

  it('should process successful openTunnelResp, when new tunnel is created', () => {
    const openTunnelRespToProcess: OpenTunnelResponseBody = {
      expiresAt: Math.floor(Date.now() / 1000) + 12 * 60 * 60,
      newTunnelCreated: true,
      sourceAccessToken: 'processOpenTunnelData-test-source-access-token',
      txId: 'processOpenTunnelData-test-tx-id',
      tunnelId: 'processOpenTunnelData-test-tunnel-id',
    };
    const expectedChargeOccurredMsg = 'Charge occurred for opening new tunnel';
    const expectedExpiresAtString = `, valid until: ${new Date(
      openTunnelRespToProcess.expiresAt * 1000,
    ).toLocaleString()}`;
    const expectedMsgToDisplay = `${expectedChargeOccurredMsg}: ${openTunnelRespToProcess.tunnelId}${expectedExpiresAtString}\n
${expectedCommonMessageToDisplay}`;

    const result = processOpenTunnelData(openTunnelRespToProcess);

    expect(result.expiresAt).toBe(openTunnelRespToProcess.expiresAt);
    expect(result.sourceAccessToken).toBe(openTunnelRespToProcess.sourceAccessToken);
    expect(result.tunnelId).toBe(openTunnelRespToProcess.tunnelId);
    expect(result.sshPort).toBe(0);
    expect(echo).toBeCalledTimes(1);
    expect(echo).toHaveBeenCalledWith(
      chalk.yellowBright(
        boxen(expectedMsgToDisplay, {
          borderStyle: 'round',
          textAlignment: 'left',
          padding: 1,
        }),
      ),
    );
  });

  it('should process successful openTunnelResp, when tunnel is reused', () => {
    const openTunnelRespToProcess = {
      expiresAt: Math.floor(Date.now() / 1000) + 11 * 60 * 60,
      newTunnelCreated: false,
      sourceAccessToken: 'processOpenTunnelData-test-source-access-token',
      txId: 'processOpenTunnelData-test-tx-id',
      tunnelId: 'processOpenTunnelData-test-tunnel-id',
    };
    const expectedChargeOccurredMsg = 'Re-using previously open tunnel';
    const expectedExpiresAtString = `, valid until: ${new Date(
      openTunnelRespToProcess.expiresAt * 1000,
    ).toLocaleString()}`;
    const expectedMsgToDisplay = `${expectedChargeOccurredMsg}: ${openTunnelRespToProcess.tunnelId}${expectedExpiresAtString}\n
${expectedCommonMessageToDisplay}`;

    const result = processOpenTunnelData(openTunnelRespToProcess);

    expect(result.expiresAt).toBe(openTunnelRespToProcess.expiresAt);
    expect(result.sourceAccessToken).toBe(openTunnelRespToProcess.sourceAccessToken);
    expect(result.tunnelId).toBe(openTunnelRespToProcess.tunnelId);
    expect(result.sshPort).toBe(0);
    expect(echo).toBeCalledTimes(1);
    expect(echo).toHaveBeenCalledWith(
      chalk.greenBright(
        boxen(expectedMsgToDisplay, {
          borderStyle: 'round',
          textAlignment: 'left',
          padding: 1,
        }),
      ),
    );
  });

  it('should process successful openTunnelResp, when expiresAt is not set', () => {
    const openTunnelRespToProcess = {
      expiresAt: 0,
      newTunnelCreated: false,
      sourceAccessToken: 'processOpenTunnelData-test-source-access-token',
      txId: 'processOpenTunnelData-test-tx-id',
      tunnelId: 'processOpenTunnelData-test-tunnel-id',
    };
    const expectedChargeOccurredMsg = 'Re-using previously open tunnel';
    const expectedExpiresAtString = '';
    const expectedMsgToDisplay = `${expectedChargeOccurredMsg}: ${openTunnelRespToProcess.tunnelId}${expectedExpiresAtString}\n
${expectedCommonMessageToDisplay}`;

    const result = processOpenTunnelData(openTunnelRespToProcess);

    expect(result.expiresAt).toBe(openTunnelRespToProcess.expiresAt);
    expect(result.sourceAccessToken).toBe(openTunnelRespToProcess.sourceAccessToken);
    expect(result.tunnelId).toBe(openTunnelRespToProcess.tunnelId);
    expect(result.sshPort).toBe(0);
    expect(echo).toBeCalledTimes(1);
    expect(echo).toHaveBeenCalledWith(
      chalk.greenBright(
        boxen(expectedMsgToDisplay, {
          borderStyle: 'round',
          textAlignment: 'left',
          padding: 1,
        }),
      ),
    );
  });

  it('should process unsuccessful openTunnelResp, when sourceAccessToken is not available and tunnelUsed info not available', () => {
    const openTunnelRespToProcess = {
      expiresAt: 0,
      newTunnelCreated: false,
      sourceAccessToken: '',
      txId: 'processOpenTunnelData-test-tx-id',
      tunnelId: 'processOpenTunnelData-test-tunnel-id',
    };
    const expectedMsgToDisplay = 'Failed to open Secure Tunnel';

    const result = processOpenTunnelData(openTunnelRespToProcess);

    expect(result.expiresAt).toBe(openTunnelRespToProcess.expiresAt);
    expect(result.sourceAccessToken).toBe(openTunnelRespToProcess.sourceAccessToken);
    expect(result.tunnelId).toBe('');
    expect(result.sshPort).toBe(0);
    expect(echo).toBeCalledTimes(1);
    expect(echo).toHaveBeenCalledWith(
      chalk.redBright(
        boxen(expectedMsgToDisplay, {
          borderStyle: 'round',
          textAlignment: 'left',
          padding: 1,
        }),
      ),
    );
  });

  test.each(testCasesForUsedBy)(
    'should process unsuccessful openTunnelResp received, and used by info needs to be displayed',
    (usedBy, sinceEpochTime, expectedMessageAddition) => {
      const openTunnelRespToProcess = {
        expiresAt: Math.floor(Date.now() / 1000) + 11 * 60 * 60,
        newTunnelCreated: false,
        sourceAccessToken: '',
        txId: 'processOpenTunnelData-test-tx-id',
        tunnelId: 'processOpenTunnelData-test-tunnel-id',
        tunnelUsed: {
          by: usedBy,
          since: sinceEpochTime,
        },
      };
      const expectedExpiresAtString = `, valid until: ${new Date(
        openTunnelRespToProcess.expiresAt * 1000,
      ).toLocaleString()}`;
      let expectedMsgToDisplay = 'Failed to open Secure Tunnel';
      expectedMsgToDisplay += `\n\n${expectedMessageAddition}${expectedExpiresAtString}`;

      const result = processOpenTunnelData(openTunnelRespToProcess);

      expect(result.expiresAt).toBe(0);
      expect(result.sourceAccessToken).toBe(openTunnelRespToProcess.sourceAccessToken);
      expect(result.tunnelId).toBe('');
      expect(result.sshPort).toBe(0);
      expect(echo).toBeCalledTimes(1);
      expect(echo).toHaveBeenCalledWith(
        chalk.redBright(
          boxen(expectedMsgToDisplay, {
            borderStyle: 'round',
            textAlignment: 'left',
            padding: 1,
          }),
        ),
      );
    },
  );
});
